from __future__ import absolute_import
# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.

import numpy as np
import scipy.optimize as opt
from . import equation_of_state as eos
from ..tools.math import bracket
import warnings


def bulk_modulus(volume, params):
    """
    compute the bulk modulus as per the third order
    birch-murnaghan equation of state.  Returns bulk
    modulus in the same units as the reference bulk
    modulus.  Pressure must be in :math:`[Pa]`.
    """

    x = params['V_0'] / volume
    f = 0.5 * (pow(x, 2. / 3.) - 1.0)

    K = pow(1. + 2. * f, 5. / 2.) * (params['K_0'] + (3. * params['K_0'] * params['Kprime_0'] -
                                                      5 * params['K_0']) * f + 27. / 2. * (params['K_0'] * params['Kprime_0'] - 4. * params['K_0']) * f * f)
    return K


def birch_murnaghan(x, params):
    """
    equation for the third order birch-murnaghan equation of state, returns
    pressure in the same units that are supplied for the reference bulk
    modulus (params['K_0'])
    """

    return 3. * params['K_0'] / 2. * (pow(x, 7. / 3.) - pow(x, 5. / 3.)) \
        * (1. - .75 * (4. - params['Kprime_0']) * (pow(x, 2. / 3.) - 1.)) + params['P_0']


def volume(pressure, params):
    """
    Get the birch-murnaghan volume at a reference temperature for a given
    pressure :math:`[Pa]`. Returns molar volume in :math:`[m^3]`
    """

    func = lambda x: birch_murnaghan(params['V_0'] / x, params) - pressure
    try:
        sol = bracket(func, params['V_0'], 1.e-2 * params['V_0'])
    except:
        raise ValueError(
            'Cannot find a volume, perhaps you are outside of the range of validity for the equation of state?')
    return opt.brentq(func, sol[0], sol[1])


def shear_modulus_second_order(volume, params):
    """
    Get the birch murnaghan shear modulus at a reference temperature, for a
    given volume.  Returns shear modulus in :math:`[Pa]` (the same units as in
    params['G_0']).  This uses a second order finite strain expansion
    """

    x = params['V_0'] / volume
    G = params['G_0'] * pow(x, 5. / 3.) * (
        1. - 0.5 * (pow(x, 2. / 3.) - 1.) * (5. - 3. * params['Gprime_0'] * params['K_0'] / params['G_0']))
    return G


def shear_modulus_third_order(volume, params):
    """
    Get the birch murnaghan shear modulus at a reference temperature, for a
    given volume.  Returns shear modulus in :math:`[Pa]` (the same units as in
    params['G_0']).  This uses a third order finite strain expansion
    """

    x = params['V_0'] / volume
    f = 0.5 * (pow(x, 2. / 3.) - 1.0)
    G = pow((1. + 2. * f), 5. / 2.) * (params['G_0'] + (3. * params['K_0'] * params['Gprime_0'] - 5. * params['G_0']) * f + (
        6. * params['K_0'] * params['Gprime_0'] - 24. * params['K_0'] - 14. * params['G_0'] + 9. / 2. * params['K_0'] * params['Kprime_0']) * f * f)
    return G


class BirchMurnaghanBase(eos.EquationOfState):

    """
    Base class for the isothermal Birch Murnaghan equation of state.  This is third order in strain, and
    has no temperature dependence.  However, the shear modulus is sometimes fit to a second order
    function, so if this is the case, you should use that.  For more see :class:`burnman.birch_murnaghan.BM2` and :class:`burnman.birch_murnaghan.BM3`.
    """

    def volume(self, pressure, temperature, params):
        """
        Returns volume :math:`[m^3]` as a function of pressure :math:`[Pa]`.
        """
        return volume(pressure, params)

    def pressure(self, temperature, volume, params):
        return birch_murnaghan(params['V_0'] / volume, params)

    def isothermal_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns isothermal bulk modulus :math:`K_T` :math:`[Pa]` as a function of pressure :math:`[Pa]`,
        temperature :math:`[K]` and volume :math:`[m^3]`.
        """
        return bulk_modulus(volume, params)

    def adiabatic_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns adiabatic bulk modulus :math:`K_s` of the mineral. :math:`[Pa]`.
        """
        return bulk_modulus(volume, params)

    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Returns shear modulus :math:`G` of the mineral. :math:`[Pa]`
        """
        if(self.order == 2):
            return shear_modulus_second_order(volume, params)
        elif(self.order == 3):
            return shear_modulus_third_order(volume, params)

    def entropy(self, pressure, temperature, volume, params):
        """
        Returns the molar entropy :math:`\mathcal{S}` of the mineral. :math:`[J/K/mol]`
        """
        return 0.

    def molar_internal_energy(self, pressure, temperature, volume, params):
        """
        Returns the internal energy :math:`\mathcal{E}` of the mineral. :math:`[J/mol]`
        """
        x = np.power(volume/params['V_0'], -1./3.)
        x2 = x*x
        x4 = x2*x2
        x6 = x4*x2
        x8 = x4*x4

        xi1 = 3.*(4. - params['Kprime_0'])/4.

        intPdV = (-9./2. * params['V_0'] * params['K_0'] *
                  ((xi1 + 1.)*(x4/4. - x2/2. + 1./4.) -
                   xi1*(x6/6. - x4/4. + 1./12.)))

        return - intPdV + params['E_0']

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the Gibbs free energy :math:`\mathcal{G}` of the mineral. :math:`[J/mol]`
        """
        # G = int VdP = [PV] - int PdV = E + PV

        return self.molar_internal_energy(pressure, temperature, volume, params) + volume*pressure

    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return a very large number. :math:`[J/K/mol]`
        """
        return 1.e99

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return a very large number. :math:`[J/K/mol]`
        """
        return 1.e99

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return zero. :math:`[1/K]`
        """
        return 0.

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return zero. :math:`[unitless]`
        """
        return 0.

    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters
        """

        if 'E_0' not in params:
            params['E_0'] = 0.
        if 'P_0' not in params:
            params['P_0'] = 0.

        # If G and Gprime are not included this is presumably deliberate,
        # as we can model density and bulk modulus just fine without them,
        # so just add them to the dictionary as nans
        if 'G_0' not in params:
            params['G_0'] = float('nan')
        if 'Gprime_0' not in params:
            params['Gprime_0'] = float('nan')

        # Check that all the required keys are in the dictionary
        expected_keys = ['V_0', 'K_0', 'Kprime_0', 'G_0', 'Gprime_0']
        for k in expected_keys:
            if k not in params:
                raise KeyError('params object missing parameter : ' + k)

        # Finally, check that the values are reasonable.
        if params['P_0'] < 0.:
            warnings.warn('Unusual value for P_0', stacklevel=2)
        if params['V_0'] < 1.e-7 or params['V_0'] > 1.e-3:
            warnings.warn('Unusual value for V_0', stacklevel=2)
        if params['K_0'] < 1.e9 or params['K_0'] > 1.e13:
            warnings.warn('Unusual value for K_0', stacklevel=2)
        if params['Kprime_0'] < 0. or params['Kprime_0'] > 10.:
            warnings.warn('Unusual value for Kprime_0', stacklevel=2)
        if params['G_0'] < 0.0 or params['G_0'] > 1.e13:
            warnings.warn('Unusual value for G_0', stacklevel=2)
        if params['Gprime_0'] < -5. or params['Gprime_0'] > 10.:
            warnings.warn('Unusual value for Gprime_0', stacklevel=2)


class BM3(BirchMurnaghanBase):

    """
    Third order Birch Murnaghan isothermal equation of state.
    This uses the third order expansion for shear modulus.
    """

    def __init__(self):
        self.order = 3


class BM2(BirchMurnaghanBase):

    """
    Third order Birch Murnaghan isothermal equation of state.
    This uses the second order expansion for shear modulus.
    """

    def __init__(self):
        self.order = 2



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import

import warnings
import numpy as np

from . import equation_of_state as eos


def tait_constants(params):
    """
    returns parameters for the modified Tait equation of state
    derived from K_T and its two first pressure derivatives
    EQ 4 from Holland and Powell, 2011
    """
    a = (1. + params['Kprime_0']) / (
        1. + params['Kprime_0'] + params['K_0'] * params['Kdprime_0'])
    b = params['Kprime_0'] / params['K_0'] - \
        params['Kdprime_0'] / (1. + params['Kprime_0'])
    c = (1. + params['Kprime_0'] + params['K_0'] * params['Kdprime_0']) / (
        params['Kprime_0'] * params['Kprime_0'] + params['Kprime_0'] - params['K_0'] * params['Kdprime_0'])
    return a, b, c


def modified_tait(x, params):
    """
    equation for the modified Tait equation of state, returns
    pressure in the same units that are supplied for the reference bulk
    modulus (params['K_0'])
    EQ 2 from Holland and Powell, 2011
    """
    a, b, c = tait_constants(params)
    return (np.power((x + a - 1.) / a, -1. / c) - 1.) / b + params['P_0']


def volume(pressure, params):
    """
    Returns volume [m^3] as a function of pressure [Pa] and temperature [K]
    EQ 12
    """
    a, b, c = tait_constants(params)
    x = 1 - a * \
        (1. - np.power((1. + b * (pressure - params['P_0'])), -1.0 * c))
    return x * params['V_0']


def bulk_modulus(pressure, params):
    """
    Returns isothermal bulk modulus :math:`K_T` of the mineral. :math:`[Pa]`.
    EQ 13+2
    """
    a, b, c = tait_constants(params)
    return params['K_0'] * (1. + b * (pressure - params['P_0'])) * (a + (1. - a) * np.power((1. + b * (pressure - params['P_0'])), c))


def intVdP(pressure, params):
    """
    Returns the integral of VdP for the mineral. :math:`[J]`.
    EQ 13
    """
    a, b, c = tait_constants(params)
    psubpth = pressure - params['P_0']

    if pressure != params['P_0']:
        intVdP = ((pressure - params['P_0'])
                  * params['V_0']
                  * (1. - a + (a * (1. - np.power((1. + b * (psubpth)), 1. - c))
                               / (b * (c - 1.)
                                  * (pressure - params['P_0'])))))
    else:
        intVdP = 0.
    return intVdP

class MT(eos.EquationOfState):

    """
    Base class for the generic modified Tait equation of state.
    References for this can be found in :cite:`HC1974`
    and :cite:`HP2011` (followed here).

    An instance "m" of a Mineral can be assigned this
    equation of state with the command m.set_method('mt')
    (or by initialising the class with the param
    equation_of_state = 'mt').
    """

    def volume(self, pressure, temperature, params):
        """
        Returns volume :math:`[m^3]` as a function of pressure :math:`[Pa]`.
        """
        return volume(pressure, params)

    def pressure(self, temperature, volume, params):
        """
        Returns pressure [Pa] as a function of temperature [K] and volume[m^3]
        """
        return modified_tait(params['V_0'] / volume, params)

    def isothermal_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns isothermal bulk modulus :math:`K_T` of the mineral. :math:`[Pa]`.
        """
        return bulk_modulus(pressure, params)

    def adiabatic_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return a very large number. :math:`[Pa]`
        """
        return 1.e99

    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Not implemented in the Modified Tait EoS. :math:`[Pa]`
        Returns 0.
        Could potentially apply a fixed Poissons ratio as a rough estimate.
        """
        return 0.

    def entropy(self, pressure, temperature, volume, params):
        """
        Returns the molar entropy :math:`\mathcal{S}` of the mineral. :math:`[J/K/mol]`
        """
        return 0.

    def molar_internal_energy(self, pressure, temperature, volume, params):
        """
        Returns the internal energy :math:`\mathcal{E}` of the mineral. :math:`[J/mol]`
        """

        return self.gibbs_free_energy(pressure, temperature, volume, params) - volume*pressure

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the Gibbs free energy :math:`\mathcal{G}` of the mineral. :math:`[J/mol]`
        """
        # G = int VdP = [PV] - int PdV = E + PV
        a, b, c = tait_constants(params)

        intVdP = params['V_0']*( a/(b*(1. - c)) *
                                 (np.power(b*(pressure - params['P_0']) + 1.,
                                           1. - c) - 1.) +
                                 (1. - a)*(pressure - params['P_0']))

        return intVdP + params['E_0'] + params['V_0']*params['P_0']

    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return a very large number. :math:`[J/K/mol]`
        """
        return 1.e99

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return a very large number. :math:`[J/K/mol]`
        """
        return 1.e99

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return zero. :math:`[1/K]`
        """
        return 0.

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return zero. :math:`[unitless]`
        """
        return 0.

    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters
        """

        if 'E_0' not in params:
            params['E_0'] = 0.
        if 'P_0' not in params:
            params['P_0'] = 1.e5

        # G and Gprime are not defined in this equation of state,
        # We can model density and bulk modulus just fine without them,
        # so just add them to the dictionary as nans
        if 'G_0' not in params:
            params['G_0'] = float('nan')
        if 'Gprime_0' not in params:
            params['Gprime_0'] = float('nan')

        # Check that all the required keys are in the dictionary
        expected_keys = [
            'V_0', 'K_0', 'Kprime_0', 'Kdprime_0', 'G_0', 'Gprime_0']
        for k in expected_keys:
            if k not in params:
                raise KeyError('params object missing parameter : ' + k)

        # Finally, check that the values are reasonable.
        if params['P_0'] < 0.:
            warnings.warn('Unusual value for P_0', stacklevel=2)
        if params['V_0'] < 1.e-7 or params['V_0'] > 1.e-2:
            warnings.warn('Unusual value for V_0', stacklevel=2)
        if params['K_0'] < 1.e9 or params['K_0'] > 1.e13:
            warnings.warn('Unusual value for K_0', stacklevel=2)
        if params['Kprime_0'] < 0. or params['Kprime_0'] > 40.:
            warnings.warn('Unusual value for Kprime_0', stacklevel=2)
        if params['G_0'] < 0.0 or params['G_0'] > 1.e13:
            warnings.warn('Unusual value for G_0', stacklevel=2)
        if params['Gprime_0'] < -5. or params['Gprime_0'] > 10.:
            warnings.warn('Unusual value for Gprime_0', stacklevel=2)



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2015 by the BurnMan team, released under the GNU GPL v2 or later.

from __future__ import absolute_import

import numpy as np
from scipy.optimize import brentq
import warnings

from . import equation_of_state as eos
from ..constants import gas_constant

class AA(eos.EquationOfState):
    """
    Class for the :math`E-V-S` liquid metal EOS detailed in :cite:`AA1994`.
    Internal energy (:math:`E`) is first calculated
    along a reference isentrope using a fourth order BM EoS
    (:math:`V_0`, :math:`KS`, :math:`KS'`, :math:`KS''`),
    which gives volume as a function of pressure,
    coupled with the thermodynamic identity:

    :math:`-\partial E/ \partial V |_S = P`.

    The temperature along the isentrope is calculated via

    :math:`\partial (\ln T)/\partial (\ln \\rho) |_S = \gamma`

    which gives:

    :math:`T_S/T_0 = \exp(\int( \gamma/\\rho ) d \\rho)`

    The thermal effect on internal energy is calculated at constant volume
    using expressions for the kinetic, electronic and potential contributions
    to the volumetric heat capacity, which can then be integrated with respect
    to temperature:

    :math:`\partial E/\partial T |_V = C_V`

    :math:`\partial E/\partial S |_V = T`

    We note that :cite:`AA1994` also include a detailed description
    of the Gruneisen parameter as a function of volume and energy (Equation 15),
    and use this to determine the temperature along the principal isentrope
    (Equations B1-B10) and the thermal pressure away from that isentrope
    (Equation 23). However, this expression is inconsistent with
    the equation of state away from the principal isentrope. Here we choose
    to calculate the thermal pressure and Grueneisen parameter thus:

    1) As energy and entropy are defined by the equation of state at any
    temperature and volume, pressure can be found by via the expression:

    :math:`\partial E/\partial V |_S = P`

    2) The Grueneisen parameter can now be determined as
    :math:`\gamma = V \partial P/\partial E |_V`

    To reiterate: away from the reference isentrope, the Grueneisen parameter
    calculated using these expressions is *not* equal to the
    (thermodynamically inconsistent) analytical expression given by :cite:`AA1994`.

    A final note: the expression for :math:`\Lambda` (Equation 17).
    does not reproduce Figure 5. We assume here that the figure matches the model
    actually used by :cite:`AA1994`, which has the form:
    :math:`F(-325.23 + 302.07 (\\rho/\\rho_0) + 30.45 (\\rho/\\rho_0)^{0.4})`.
    """

    def _ABTheta(self, V, params):
        """
        Electronic heat capacity functions
        """
        Vfrac = V/params['V_0']

        A = params['a'][0] + params['a'][1]*Vfrac # A2
        B = params['b'][0] + params['b'][1]*Vfrac*Vfrac # A3
        Theta = params['Theta'][0]*np.power(Vfrac, -params['Theta'][1]) # A4

        return A, B, Theta

    def _lambdaxi(self, V, params):
        """
        Potential heat capacity functions
        """
        rhofrac = params['V_0']/V
        xi = params['xi_0']*np.power(rhofrac, -0.6) # A16
        F = 1./(1. + np.exp((rhofrac - params['F'][0])/params['F'][1])) # A18
        #lmda = (F*(params['lmda'][0] + params['lmda'][1]*rhofrac) + params['lmda'][2])*np.power(rhofrac, 0.4) # A17
        lmda = (F*(params['lmda'][0] + params['lmda'][1]*rhofrac + params['lmda'][2]*np.power(rhofrac, 0.4))) # this incorrect expression for lmda seems to provide a very close fit to figure 5

        return lmda, xi

    def _rhofracxksis(self, V, params):
        """
        Functions for the fourth order Birch-Murnaghan equation of state
        """
        rhofrac = params['V_0']/V # rho/rho0 = V0/V
        x = np.power(rhofrac, 1./3.) # equation 18
        ksi1 = 0.75*(4. - params['Kprime_S']) # equation 19
        ksi2 = 0.375*(params['K_S']*params['Kprime_prime_S'] +
                      params['Kprime_S']*(params['Kprime_S'] - 7.)) + 143./24. # equation 20
        return rhofrac, x, ksi1, ksi2


    def _isentropic_temperature(self, V, params):
        """
        Temperature along the reference isentrope
        """

        rhofrac, x, ksi1, ksi2 = self._rhofracxksis(V, params)

        # equation B6 -- B10
        a1 = ksi2 / 8.
        a2 = ( ksi1 + 3. * ksi2 ) / 6.
        a3 = ( 1. + 2.*ksi1 + 3.*ksi2 ) / 4.
        a4 = (1. + ksi1 + ksi2)/2.
        a5 = (6. + 4.*ksi1 + 3.*ksi2)/24.

        # equation B5
        Ts = params['T_0']*np.exp(params['grueneisen_0']*np.log(rhofrac)
                                + 13.5*params['grueneisen_prime']*params['V_0']*params['K_S'] *
                                (   (a1/(3*params['grueneisen_n'] + 8.))*(np.power(x,(3*params['grueneisen_n'] + 8.)) - 1.)
                                    - (a2/(3*params['grueneisen_n'] + 6.))*(np.power(x,(3*params['grueneisen_n'] + 6.)) - 1.)
                                + (a3/(3*params['grueneisen_n'] + 4.))*(np.power(x,(3*params['grueneisen_n'] + 4.)) - 1.)
                                - (a4/(3*params['grueneisen_n'] + 2.))*(np.power(x,(3*params['grueneisen_n'] + 2.)) - 1.)
                                + (a5/(3*params['grueneisen_n'] + 0.))*(np.power(x,(3*params['grueneisen_n'] + 0.)) - 1.)))

        return Ts



    def _isentropic_pressure(self, V, params):
        """
        Pressure along the reference isentrope
        """
        rhofrac, x, ksi1, ksi2 = self._rhofracxksis(V, params)
        x2 = x*x
        x3 = x*x*x
        x5 = x3*x2
        x7 = x5*x2

        Ps = ( 1.5*params['K_S'] * (x7 - x5) *
               (1. + ksi1 - ksi1*x2 +
                ksi2 * (x2 - 1.) * (x2 - 1.)) ) # Eq. 17

        return Ps

    def _isentropic_energy_change(self, V, params):
        """
        Birch Murnaghan equation of state expression for the energy change along an isentrope
        """
        rhofrac, x, ksi1, ksi2 = self._rhofracxksis(V, params)
        x2 = x*x
        x4 = x2*x2
        x6 = x4*x2
        x8 = x4*x4

        E_S = 4.5*params['V_0']*params['K_S'] * ((ksi1 + 1.) * (x4/4. - x2/2. + 1./4.) -
                                                 ksi1*(x6/6. - x4/4. + 1./12.) +
                                                 ksi2*(x8/8. - x6/2. + 3.*x4/4. - x2/2. + 1./8.)) # Eq. 21
        return E_S

    def _isochoric_energy_change(self, Ts, T, V, params):
        """
        int Cv dT
        """
        A, B, Theta = self._ABTheta(V, params)
        lmda, xi = self._lambdaxi(V, params)

        E_kin = 1.5*params['n']*gas_constant*(T - Ts)
        E_el = A*(T - Ts - Theta*(np.arctan(T/Theta) - np.arctan(Ts/Theta))) + 5./8*B*(np.power(T, 1.6) - np.power(Ts, 1.6)) # A5
        E_pot = (lmda*(T - Ts) + params['theta']*(xi - lmda)*np.log((params['theta'] + T)/(params['theta'] + Ts))) # A19

        return E_kin + E_el + E_pot


    def volume_dependent_q(self, x, params):
        """
        Finite strain approximation for :math:`q`, the isotropic volume strain
        derivative of the grueneisen parameter.
        """
        raise NotImplementedError("")

    def _isotropic_eta_s(self, x, params):
        """
        Finite strain approximation for :math:`eta_{s0}`, the isotropic shear
        strain derivative of the grueneisen parameter.
        Zero for a liquid
        """
        return 0.

    def volume(self, pressure, temperature, params):
        """
        Returns molar volume. :math:`[m^3]`
        """

        _volume = lambda V, P, T, params: ( P -
                                            self.pressure(T, V, params) )

        return brentq(_volume, params['V_0']*0.1, params['V_0']*2., args=(pressure, temperature, params))

    def pressure( self, temperature, volume, params):
        """
        Returns the pressure of the mineral at a given temperature and volume [Pa]
        """

        '''
        Ts = self._isentropic_temperature(volume, params)


        dE = self._isochoric_energy_change(Ts, temperature, volume, params)
        E1 = self._isentropic_energy_change(volume, params) - params['E_0']
        E2 = E1 + dE

        # Integrate at constant volume (V \int dP = \int gr dE)
        dP = (params['grueneisen_0']*(E2 - E1) +
              (0.5*params['grueneisen_prime'] *
               np.power(params['V_0']/volume, params['grueneisen_n']) *
               (E2*E2 - E1*E1))) / volume # eq. 23

        P = self._isentropic_pressure(volume, params) + dP
        '''

        dV = volume*1.e-4
        S = self.entropy(0., temperature, volume, params)

        delta_S = lambda T, S, V: S - self.entropy(0., T, V, params)

        T0 = brentq(delta_S, temperature*0.97, temperature*1.03, args=(S, volume - 0.5*dV))
        T1 = brentq(delta_S, temperature*0.97, temperature*1.03, args=(S, volume + 0.5*dV))

        E0 = self.molar_internal_energy(0., T0, volume - 0.5*dV, params)
        E1 = self.molar_internal_energy(0., T1, volume + 0.5*dV, params)

        P = -(E1 - E0)/dV # |S

        return P



    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Returns grueneisen parameter :math:`[unitless]`
        """
        '''
        gr = (params['grueneisen_0'] +
              params['grueneisen_prime'] *
              (np.power(params['V_0']/volume, params['grueneisen_n']) *
               (self.molar_internal_energy(pressure, temperature, volume, params) -
                params['E_0'])))
        '''
        dT = 1.
        dE = (self.molar_internal_energy(0., temperature + 0.5*dT, volume, params) -
              self.molar_internal_energy(0., temperature - 0.5*dT, volume, params))
        dP = (self.pressure(temperature + 0.5*dT, volume, params) -
              self.pressure(temperature - 0.5*dT, volume, params))
        gr = volume*dP/dE

        return gr

    def isothermal_bulk_modulus(self, pressure,temperature, volume, params):
        """
        Returns isothermal bulk modulus :math:`[Pa]`
        """
        # K_T = -V * dP/dV
        dV = volume*1.e-3
        P0 = self.pressure(temperature, volume - 0.5*dV, params)
        P1 = self.pressure(temperature, volume + 0.5*dV, params)

        K_T = -volume*(P1 - P0)/dV
        return K_T

    def adiabatic_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns adiabatic bulk modulus. :math:`[Pa]`
        """
        K_T=self.isothermal_bulk_modulus(pressure, temperature, volume, params)
        alpha = self.thermal_expansivity(pressure, temperature, volume, params)
        gr = self.grueneisen_parameter(pressure, temperature, volume, params)
        K_S = K_T*(1. + gr * alpha * temperature)
        return K_S

    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Returns shear modulus. :math:`[Pa]`
        Zero for a liquid
        """
        return 0.

    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Returns heat capacity at constant volume. :math:`[J/K/mol]`
        """

        A, B, Theta = self._ABTheta(volume, params)
        lmda, xi = self._lambdaxi(volume, params)

        C_kin = 1.5*params['n']*gas_constant # HT limit of kinetic contribution (just after equation 29.)
        C_e = A*(1. - (Theta*Theta)/(Theta*Theta + temperature*temperature)) + B*np.power(temperature, 0.6) # Equation A1
        C_pot = (lmda*temperature + xi*params['theta']) / (params['theta'] + temperature) # Equation A15

        return C_kin + C_e + C_pot


    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Returns heat capacity at constant pressure. :math:`[J/K/mol]`
        """

        alpha = self.thermal_expansivity(pressure, temperature, volume, params)
        gr = self.grueneisen_parameter(pressure, temperature, volume, params)
        C_v = self.molar_heat_capacity_v(pressure, temperature, volume, params)
        C_p = C_v*(1. + gr * alpha * temperature)

        return C_p

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Returns thermal expansivity. :math:`[1/K]`
        Currently found by numerical differentiation (1/V * dV/dT)
        """

        delta_T = 1.
        V0 = self.volume(pressure, temperature-0.5*delta_T, params)
        V1 = self.volume(pressure, temperature+0.5*delta_T, params)

        return (1./volume)*(V1 - V0)/delta_T

    def gibbs_free_energy( self, pressure, temperature, volume, params):
        """
        Returns the Gibbs free energy at the pressure and temperature of the mineral [J/mol]
        E + PV
        """

        return self.helmholtz_free_energy( pressure, temperature, volume, params) + \
            pressure * self.volume( pressure, temperature, params)

    def molar_internal_energy( self, pressure, temperature, volume, params):
        """
        Returns the internal energy at the pressure and temperature of the mineral [J/mol]
        """
        Ts = self._isentropic_temperature(volume, params)
        E = (params['E_0'] + self._isentropic_energy_change(volume, params)
                + self._isochoric_energy_change(Ts, temperature, volume, params))

        return E

    def entropy( self, pressure, temperature, volume, params):
        """
        Returns the entropy at the pressure and temperature of the mineral [J/K/mol]
        """
        T = temperature
        Ts = self._isentropic_temperature(volume, params)

        if np.abs(T- Ts) < 1.e-10:
            Delta_S = 0.
        else:
            A, B, Theta = self._ABTheta(volume, params)
            lmda, xi = self._lambdaxi(volume, params)
            S_kin = 1.5*params['n']*gas_constant*(np.log(T) - np.log(Ts))
            S_el = (A*(np.log(T/Ts) - 0.5*np.log(T*T*(Theta*Theta + Ts*Ts)/(Ts*Ts*(Theta*Theta + T*T)))) + 5./3.*B*(np.power(T, 0.6) - np.power(Ts, 0.6))) # A6
            S_pot = (lmda*np.log((params['theta'] + T)/(params['theta'] + Ts)) + xi*np.log((T*(params['theta'] + Ts))/(Ts*(params['theta'] + T)))) # A20
            Delta_S = S_kin + S_el + S_pot

        S = params['S_0'] + Delta_S
        return S

    def enthalpy( self, pressure, temperature, volume, params):
        """
        Returns the enthalpy at the pressure and temperature of the mineral [J/mol]
        E + PV
        """

        return self.molar_internal_energy(pressure, temperature, volume, params) + \
            pressure * self.volume( pressure, temperature, params)

    def helmholtz_free_energy( self, pressure, temperature, volume, params):
        """
        Returns the Helmholtz free energy at the pressure and temperature of the mineral [J/mol]
        E - TS
        """
        return self.molar_internal_energy(pressure, temperature, volume, params) - temperature*self.entropy(pressure, temperature, volume, params)


    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters
        """

        # Now check all the required keys for the
        # thermal part of the EoS are in the dictionary
        expected_keys = ['P_0', 'T_0', 'S_0', 'molar_mass', 'grueneisen_0']


        for k in expected_keys:
            if k not in params:
                raise KeyError('params object missing parameter : ' + k)

        # Finally, check that the values are reasonable.
        if params['T_0'] < 0.:
            warnings.warn( 'Unusual value for T_0', stacklevel=2 )
        if params['molar_mass'] < 0.001 or params['molar_mass'] > 10.:
            warnings.warn( 'Unusual value for molar_mass', stacklevel=2 )
        if params['n'] < 1. or params['n'] > 1000.:
            warnings.warn( 'Unusual value for n', stacklevel=2 )



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit
# for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

from __future__ import absolute_import

import importlib
import numpy as np
from sympy import Rational
from fractions import Fraction
from scipy.spatial import Delaunay
from scipy.special import comb
from copy import copy

from .material import cached_property

from ..tools.math import independent_row_indices

try:
    cdd = importlib.import_module('cdd')
except ImportError as err:
    print(f'Warning: {err}. '
          'For full functionality of BurnMan, please install pycddlib.')


class SimplexGrid(object):
    """
    A class that creates objects that can efficiently generate a set of points
    that grid a simplex with a user-defined number of vertices. The class
    contains both a generator method and a grid method. It also contains
    an n_points attribute that returns the number of points in the gridded
    simplex.

    This class is available as :class:`burnman.polytope.SimplexGrid`.
    """

    def __init__(self, vertices, points_per_edge):
        """
        Initialize SimplexGrid object with the desired number of vertices
        and points per edge.
        """
        assert vertices >= 2, 'need at least two vertices'
        assert points_per_edge >= 2, 'need at least 2 points per edge'

        self.vertices = vertices
        self.points_per_edge = points_per_edge

    def generate(self, generate_type='list'):
        """
        Generates the grid points of the simplex in lexicographic order.

        Parameters
        ----------
        generate_type : 'list' or 'array'
            Determines whether the generator returns lists or arrays
            corresponding to each point in the simplex grid.

        Returns
        -------
        generator of lists or ndarrays (int, ndim=1)
            Grid points of the simplex.
        """

        if generate_type == 'list':
            x = [0]*self.vertices
        elif generate_type == 'array':
            x = np.zeros(self.vertices, dtype=int)
        else:
            raise Exception('generate_type must be of type list or array.')

        x[self.vertices-1] = self.points_per_edge-1

        h = self.vertices
        while True:
            yield copy(x)

            h -= 1
            if h == 0:
                return

            val = x[h]
            x[h] = 0
            x[self.vertices-1] = val - 1
            x[h-1] += 1
            if val != 1:
                h = self.vertices

    def grid(self, generate_type='list'):
        """
        Returns either a list or a numpy array
        corresponding the the points in the simplex grid, depending on
        whether the user chooses 'list' (default) or 'array' as
        the generate_type parameter.
        """
        if generate_type == 'list':
            return list(self.generate(generate_type))
        elif generate_type == 'array':
            return np.array(list(self.generate(generate_type)))
        else:
            raise Exception('generate_type must be of type list or array.')

    def n_points(self):
        """
        The number of points corresponding to the number of vertices and
        points per edge chosen by the user.
        """
        return comb(self.vertices+self.points_per_edge-2,
                    self.vertices-1, exact=True)


class MaterialPolytope(object):
    """
    A class that can be instantiated to create pycddlib polytope objects.
    These objects can be interrogated to provide the vertices satisfying the
    input constraints.

    This class is available as :class:`burnman.polytope.MaterialPolytope`.
    """

    def __init__(self, equalities,
                 inequalities,
                 number_type='fraction',
                 return_fractions=False,
                 independent_endmember_occupancies=None):
        """
        Initialization function for the MaterialPolytope class.
        Declares basis attributes of the class.

        Parameters
        ----------
        equalities: 2D numpy array
            A numpy array containing all the equalities of the polytope.
            Each row should evaluate to 0.
        inequalities: 2D numpy array
            A numpy array containing all the inequalities of the polytope.
            Each row should evaluate to <= 0.
        number_type: 'fraction' or 'float' (default is 'fraction')
            Whether pycddlib should read the input arrays as
            fractions or floats.
        return_fractions : boolean (default is False)
            Whether the generated polytope object should return fractions or
            floats.
        independent_endmember_occupancies : 2D numpy array (or None)
            If specified, this array provides the independent endmember set
            against which the dependent endmembers are defined.
        """
        self.set_return_type(return_fractions)
        self.equality_matrix = equalities[:, 1:]
        self.equality_vector = -equalities[:, 0]

        self.polytope_matrix = cdd.Matrix(equalities, linear=True,
                                          number_type=number_type)
        self.polytope_matrix.rep_type = cdd.RepType.INEQUALITY
        self.polytope_matrix.extend(inequalities, linear=False)
        self.polytope = cdd.Polyhedron(self.polytope_matrix)

        if independent_endmember_occupancies is not None:
            self.independent_endmember_occupancies = independent_endmember_occupancies

    def set_return_type(self, return_fractions=False):
        """
        Sets the return_type for the polytope object. Also deletes the cached
        endmember_occupancies property.

        Parameters
        ----------
        return_fractions : boolean (default is False)
            Whether the generated polytope object should return fractions or
            floats.
        """
        try:
            del self.__dict__['endmember_occupancies']
        except KeyError:
            pass
        self.return_fractions = return_fractions

    @cached_property
    def raw_vertices(self):
        """
        Returns a list of the vertices of the polytope without any
        postprocessing. See also endmember_occupancies.
        """
        return self.polytope.get_generators()[:]

    @cached_property
    def limits(self):
        """
        Return the limits of the polytope (the set of bounding inequalities).
        """
        return np.array(self.polytope.get_inequalities(), dtype=float)

    @cached_property
    def n_endmembers(self):
        """
        Return the number of endmembers
        (the number of vertices of the polytope).
        """
        return len(self.raw_vertices)

    @cached_property
    def endmember_occupancies(self):
        """
        Return the endmember occupancies
        (a processed list of all of the vertex locations).
        """
        if self.return_fractions:
            if self.polytope.number_type == 'fraction':
                v = np.array([[Fraction(value) for value in v]
                              for v in self.raw_vertices])
            else:
                v = np.array([[Rational(value).limit_denominator(1000000)
                               for value in v]
                              for v in self.raw_vertices])
        else:
            v = np.array([[float(value) for value in v]
                          for v in self.raw_vertices])

        if len(v.shape) == 1:
            raise ValueError("The combined equality and positivity "
                             "constraints result in a null polytope.")

        return v[:, 1:] / v[:, 0, np.newaxis]

    @cached_property
    def independent_endmember_occupancies(self):
        """
        Return an independent set of endmember occupancies
        (a linearly-independent set of vertex locations)
        """
        arr = self.endmember_occupancies
        return arr[independent_row_indices(arr)]

    @cached_property
    def endmembers_as_independent_endmember_amounts(self):
        """
        Return a list of all the endmembers as a linear sum of
        the independent endmembers.
        """
        ind = self.independent_endmember_occupancies

        sol = np.linalg.lstsq(np.array(ind.T).astype(float),
                              np.array(self.endmember_occupancies.T).astype(
                                  float),
                              rcond=0)[0].round(decimals=12).T
        return sol

    def _decompose_vertices_into_simplices(self, vertices):
        """
        Decomposes a set of vertices into simplices by Delaunay triangulation.
        """
        # Delaunay triangulation only works in dimensions > 1
        # and we remove the nullspace (sum(fractions) = 1)
        if len(vertices) > 2:
            nulls = np.repeat(vertices[:, -1],
                              vertices.shape[1]).reshape(vertices.shape)
            tri = Delaunay((vertices - nulls)[:, :-1])
            return tri.simplices
        else:
            return [[0, 1]]

    @cached_property
    def independent_endmember_polytope(self):
        """
        Returns the polytope expressed in terms of proportions of the
        independent endmembers. The polytope involves the first
        n-1 independent endmembers. The last endmember proportion makes
        the sum equal to one.
        """
        arr = self.endmembers_as_independent_endmember_amounts
        arr = np.hstack((np.ones((len(arr), 1)), arr[:, :-1]))
        M = cdd.Matrix(arr, number_type='fraction')
        M.rep_type = cdd.RepType.GENERATOR
        return cdd.Polyhedron(M)

    @cached_property
    def independent_endmember_limits(self):
        """
        Gets the limits of the polytope as a function of the independent
        endmembers.
        """
        return np.array(self.independent_endmember_polytope.get_inequalities(),
                        dtype=float)

    def subpolytope_from_independent_endmember_limits(self, limits):
        """
        Returns a smaller polytope by applying additional limits to the amounts
        of the independent endmembers.
        """
        modified_limits = self.independent_endmember_polytope.get_inequalities().copy()
        modified_limits.extend(limits, linear=False)
        return cdd.Polyhedron(modified_limits)

    def subpolytope_from_site_occupancy_limits(self, limits):
        """
        Returns a smaller polytope by applying additional limits to the
        individual site occupancies.
        """
        modified_limits = self.polytope_matrix.copy()
        modified_limits.extend(limits, linear=False)
        return cdd.Polyhedron(modified_limits)

    def grid(self, points_per_edge=2, unique_sorted=True,
             grid_type='independent endmember proportions', limits=None):
        """
        Create a grid of points which span the polytope.

        Parameters
        ----------
        points_per_edge : integer (default is 2)
            Number of points per edge of the polytope.
        unique_sorted : boolean (default is True)
            The gridding is done by splitting the polytope into
            a set of simplices. This means that points will be duplicated along
            vertices, faces etc. If unique_sorted is True, this function
            will sort and make the points unique. This is an expensive
            operation for large polytopes, and may not always be necessary.
        grid_type : 'independent endmember proportions' (default) or 'site occupancies'
            Whether to grid the polytope in terms of
            independent endmember proportions or site occupancies.
        limits : 2D numpy array
            Additional inequalities restricting the gridded area of the polytope.
        Returns
        -------
        points : 2D numpy array
            A list of points gridding the polytope.
        """
        if limits is None:
            if grid_type == 'independent endmember proportions':
                f_occ = (self.endmembers_as_independent_endmember_amounts
                         / (points_per_edge - 1))
            elif grid_type == 'site occupancies':
                f_occ = self.endmember_occupancies/(points_per_edge-1)
            else:
                raise Exception('grid type not recognised. Should be one of '
                                'independent endmember proportions '
                                'or site occupancies')

            simplices = self._decompose_vertices_into_simplices(
                self.endmembers_as_independent_endmember_amounts)
        else:
            if grid_type == 'independent endmember proportions':
                ppns = np.array(self.subpolytope_from_independent_endmember_limits(
                    limits).get_generators()[:])[:, 1:]
                last_ppn = np.array([1. - sum(p)
                                    for p in ppns]).reshape((len(ppns), 1))
                vertices_as_independent_endmember_proportions = np.hstack(
                    (ppns, last_ppn))
                f_occ = vertices_as_independent_endmember_proportions / \
                    (points_per_edge-1)

            elif grid_type == 'site occupancies':
                occ = np.array(self.subpolytope_from_site_occupancy_limits(
                    limits).get_generators()[:])[:, 1:]
                f_occ = occ/(points_per_edge-1)

                ind = self.independent_endmember_occupancies

                vertices_as_independent_endmember_proportions = np.linalg.lstsq(np.array(ind.T).astype(float),
                                                                                np.array(occ.T).astype(
                                                                                    float),
                                                                                rcond=None)[0].round(decimals=12).T
            else:
                raise Exception('grid_type not recognised. '
                                'Should be one of '
                                'independent endmember proportions '
                                'or site occupancies')

            simplices = self._decompose_vertices_into_simplices(
                vertices_as_independent_endmember_proportions)

        n_ind = f_occ.shape[1]
        n_simplices = len(simplices)
        dim = len(simplices[0])
        simplex_grid = SimplexGrid(dim, points_per_edge)
        grid = simplex_grid.grid('array')
        points_per_simplex = simplex_grid.n_points()
        n_points = n_simplices*points_per_simplex

        points = np.empty((n_points, n_ind))
        idx = 0
        for i in range(0, n_simplices):
            points[idx:idx+points_per_simplex] = grid.dot(f_occ[simplices[i]])
            idx += points_per_simplex

        if unique_sorted:
            points = np.unique(points, axis=0)
        return points



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.


"""
SLB_2011
^^^^^^^^

Minerals from Stixrude & Lithgow-Bertelloni 2011 and references therein.
File autogenerated using SLBdata_to_burnman.py.
"""

from __future__ import absolute_import

from ..classes.mineral import Mineral
from ..classes.solidsolution import SolidSolution
from ..classes.solutionmodel import *
from ..tools.chemistry import dictionarize_formula, formula_mass

'''
SOLID SOLUTIONS
from inv251010 of HeFESTo
'''


class c2c_pyroxene(SolidSolution):

    def __init__(self, molar_fractions=None):
        self.name = 'C2/c pyroxene'
        self.solution_type = 'ideal'
        self.endmembers = [
            [hp_clinoenstatite(), '[Mg]2Si2O6'], [hp_clinoferrosilite(), '[Fe]2Si2O6']]

        SolidSolution.__init__(self, molar_fractions=molar_fractions)


class ca_ferrite_structured_phase(SolidSolution):

    def __init__(self, molar_fractions=None):
        self.name = 'calcium ferrite structured phase'
        self.solution_type = 'ideal'
        self.endmembers = [[mg_ca_ferrite(), '[Mg]Al[Al]O4'], [
                           fe_ca_ferrite(), '[Fe]Al[Al]O4'], [na_ca_ferrite(), '[Na]Al[Si]O4']]

        SolidSolution.__init__(self, molar_fractions=molar_fractions)


class clinopyroxene(SolidSolution):

    def __init__(self, molar_fractions=None):
        self.name = 'clinopyroxene'
        self.solution_type = 'asymmetric'
        self.endmembers = [[diopside(), '[Ca][Mg][Si]2O6'], [hedenbergite(), '[Ca][Fe][Si]2O6'], [
                           clinoenstatite(), '[Mg][Mg][Si]2O6'], [ca_tschermaks(), '[Ca][Al][Si1/2Al1/2]2O6'], [jadeite(), '[Na][Al][Si]2O6']]
        self.alphas = [1., 1., 1., 3.5, 1.]
        self.energy_interaction = [
            [0., 24.74e3, 26.e3, 24.3e3], [24.74e3, 0., 0.e3], [60.53136e3, 0.0], [10.e3]]

        SolidSolution.__init__(self, molar_fractions=molar_fractions)


class garnet(SolidSolution):

    def __init__(self, molar_fractions=None):
        self.name = 'garnet'
        self.solution_type = 'symmetric'
        self.endmembers = [[pyrope(), '[Mg]3[Al][Al]Si3O12'], [almandine(), '[Fe]3[Al][Al]Si3O12'], [
                                   grossular(), '[Ca]3[Al][Al]Si3O12'], [mg_majorite(), '[Mg]3[Mg][Si]Si3O12'], [jd_majorite(), '[Na2/3Al1/3]3[Al][Si]Si3O12']]
        self.energy_interaction = [
            [0.0, 30.e3, 21.20278e3, 0.0], [0.0, 0.0, 0.0], [57.77596e3, 0.0], [0.0]]

        SolidSolution.__init__(self, molar_fractions=molar_fractions)


class akimotoite(SolidSolution):

    def __init__(self, molar_fractions=None):
        self.name = 'akimotoite/ilmenite'
        self.solution_type = 'symmetric'
        self.endmembers = [[mg_akimotoite(), '[Mg][Si]O3'], [
                           fe_akimotoite(), '[Fe][Si]O3'], [corundum(), '[Al][Al]O3']]
        self.energy_interaction = [[0.0, 66.e3], [0.e3]]

        SolidSolution.__init__(self, molar_fractions=molar_fractions)


class ferropericlase(SolidSolution):

    def __init__(self, molar_fractions=None):
        self.name = 'magnesiowustite/ferropericlase'
        self.solution_type = 'symmetric'
        self.endmembers = [[periclase(), '[Mg]O'], [wuestite(), '[Fe]O']]
        self.energy_interaction = [[13.e3]]

        SolidSolution.__init__(self, molar_fractions=molar_fractions)


class mg_fe_olivine(SolidSolution):

    def __init__(self, molar_fractions=None):
        self.name = 'olivine'
        self.solution_type = 'symmetric'
        self.endmembers = [[
            forsterite(), '[Mg]2SiO4'], [fayalite(), '[Fe]2SiO4']]
        self.energy_interaction = [[7.81322e3]]

        SolidSolution.__init__(self, molar_fractions=molar_fractions)


class orthopyroxene(SolidSolution):

    def __init__(self, molar_fractions=None):
        self.name = 'orthopyroxene'
        self.solution_type = 'symmetric'
        self.endmembers = [[enstatite(), '[Mg][Mg]SiSiO6'], [ferrosilite(), '[Fe][Fe]SiSiO6'], [
                           mg_tschermaks(), '[Mg][Al]AlSiO6'], [ortho_diopside(), '[Ca][Mg]SiSiO6']] # Al avoidance, see Figure 3
        self.energy_interaction = [
            [0.0, 0.0, 32.11352e3], [0.0, 0.0], [48.35316e3]]

        SolidSolution.__init__(self, molar_fractions=molar_fractions)


class plagioclase(SolidSolution):

    def __init__(self, molar_fractions=None):
        self.name = 'plagioclase'
        self.solution_type = 'symmetric'
        self.endmembers = [
            [anorthite(), '[Ca][Al]2Si2O8'], [albite(), '[Na][Al1/2Si1/2]2Si2O8']]
        self.energy_interaction = [[26.0e3]]

        SolidSolution.__init__(self, molar_fractions=molar_fractions)


class post_perovskite(SolidSolution):

    def __init__(self, molar_fractions=None):
        self.name = 'post-perovskite/bridgmanite'
        self.solution_type = 'symmetric'
        self.endmembers = [[mg_post_perovskite(), '[Mg][Si]O3'], [
                           fe_post_perovskite(), '[Fe][Si]O3'], [al_post_perovskite(), '[Al][Al]O3']]
        self.energy_interaction = [[0.0, 60.0e3], [0.0]]

        SolidSolution.__init__(self, molar_fractions=molar_fractions)


class mg_fe_perovskite(SolidSolution):

    def __init__(self, molar_fractions=None):
        self.name = 'magnesium silicate perovskite/bridgmanite'
        self.solution_type = 'asymmetric'
        self.endmembers = [[mg_perovskite(), '[Mg][Si]O3'], [
                           fe_perovskite(), '[Fe][Si]O3'], [al_perovskite(), '[Al][Al]O3']]
        self.energy_interaction = [[0.0, 116.0e3], [0.0]]

        self.alphas = [1., 1., 0.39]
        SolidSolution.__init__(self, molar_fractions=molar_fractions)


class mg_fe_ringwoodite(SolidSolution):

    def __init__(self, molar_fractions=None):
        self.name = 'ringwoodite'
        self.solution_type = 'symmetric'
        self.endmembers = [
            [mg_ringwoodite(), '[Mg]2SiO4'], [fe_ringwoodite(), '[Fe]2SiO4']]
        self.energy_interaction = [[9.34084e3]]

        SolidSolution.__init__(self, molar_fractions=molar_fractions)


class mg_fe_aluminous_spinel(SolidSolution):

    def __init__(self, molar_fractions=None):
        self.name = 'spinel-hercynite binary, fixed order'
        self.solution_type = 'symmetric'
        self.endmembers = [[spinel(), '[Mg3/4Al1/4]4[Al7/8Mg1/8]8O16'], [
                                   hercynite(), '[Fe3/4Al1/4]4[Al7/8Fe1/8]8O16']]
        self.energy_interaction = [[5.87646e3]]

        SolidSolution.__init__(self, molar_fractions=molar_fractions)


class mg_fe_wadsleyite(SolidSolution):

    def __init__(self, molar_fractions=None):
        self.name = 'wadsleyite'
        self.solution_type = 'symmetric'
        self.endmembers = [
            [mg_wadsleyite(), '[Mg]2SiO4'], [fe_wadsleyite(), '[Fe]2SiO4']]
        self.energy_interaction = [[16.74718e3]]

        SolidSolution.__init__(self, molar_fractions=molar_fractions)

"""
ENDMEMBERS
"""


class anorthite (Mineral):

    def __init__(self):
        formula = 'CaAl2Si2O8'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Anorthite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -4014619.0,
            'V_0': 0.00010061,
            'K_0': 84089150000.0,
            'Kprime_0': 4.0,
            'Debye_0': 752.3911,
            'grueneisen_0': 0.39241,
            'q_0': 1.0,
            'G_0': 39900000000.0,
            'Gprime_0': 1.09134,
            'eta_s_0': 1.6254,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 4000.0,
            'err_V_0': 0.0,
            'err_K_0': 5000000000.0,
            'err_K_prime_0': 1.0,
            'err_Debye_0': 2.0,
            'err_grueneisen_0': 0.05,
            'err_q_0': 1.0,
            'err_G_0': 3000000000.0,
            'err_Gprime_0': 0.5,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class albite (Mineral):

    def __init__(self):
        formula = 'NaAlSi3O8'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Albite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -3718799.0,
            'V_0': 0.000100452,
            'K_0': 59761620000.0,
            'Kprime_0': 4.0,
            'Debye_0': 713.7824,
            'grueneisen_0': 0.56704,
            'q_0': 1.0,
            'G_0': 36000000000.0,
            'Gprime_0': 1.3855,
            'eta_s_0': 1.04208,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 5000.0,
            'err_V_0': 0.0,
            'err_K_0': 5000000000.0,
            'err_K_prime_0': 1.0,
            'err_Debye_0': 13.0,
            'err_grueneisen_0': 0.03,
            'err_q_0': 1.0,
            'err_G_0': 5000000000.0,
            'err_Gprime_0': 0.5,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class spinel (Mineral):

    def __init__(self):
        formula = 'Mg4Al8O16'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Spinel',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -8667568.0,
            'V_0': 0.000159048,
            'K_0': 1.969428e+11,
            'Kprime_0': 5.68282,
            'Debye_0': 842.8104,
            'grueneisen_0': 1.02283,
            'q_0': 2.71208,
            'G_0': 1.085e+11,
            'Gprime_0': 0.37303,
            'eta_s_0': 2.66282,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.property_modifiers = [
            ['linear', {'delta_E': 0.0, 'delta_S': 43.76, 'delta_V': 0.0}]]

        self.uncertainties = {
            'err_F_0': 32000.0,
            'err_V_0': 0.0,
            'err_K_0': 1000000000.0,
            'err_K_prime_0': 0.2,
            'err_Debye_0': 33.0,
            'err_grueneisen_0': 0.04,
            'err_q_0': 0.6,
            'err_G_0': 10000000000.0,
            'err_Gprime_0': 0.5,
            'err_eta_s_0': 0.6}
        Mineral.__init__(self)


class hercynite (Mineral):

    def __init__(self):
        formula = 'Fe4Al8O16'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Hercynite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -7324009.0,
            'V_0': 0.000163372,
            'K_0': 2.088965e+11,
            'Kprime_0': 5.68282,
            'Debye_0': 763.231,
            'grueneisen_0': 1.21719,
            'q_0': 2.71208,
            'G_0': 84500000000.0,
            'Gprime_0': 0.37303,
            'eta_s_0': 2.768,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.property_modifiers = [
            ['linear', {'delta_E': 0.0, 'delta_S': 97.28, 'delta_V': 0.0}]]

        self.uncertainties = {
            'err_F_0': 35000.0,
            'err_V_0': 0.0,
            'err_K_0': 2000000000.0,
            'err_K_prime_0': 1.0,
            'err_Debye_0': 32.0,
            'err_grueneisen_0': 0.07,
            'err_q_0': 1.0,
            'err_G_0': 13000000000.0,
            'err_Gprime_0': 0.5,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class forsterite (Mineral):

    def __init__(self):
        formula = 'Mg2SiO4'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Forsterite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -2055403.0,
            'V_0': 4.3603e-05,
            'K_0': 1.279555e+11,
            'Kprime_0': 4.21796,
            'Debye_0': 809.1703,
            'grueneisen_0': 0.99282,
            'q_0': 2.10672,
            'G_0': 81599990000.0,
            'Gprime_0': 1.46257,
            'eta_s_0': 2.29972,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 2000.0,
            'err_V_0': 0.0,
            'err_K_0': 2000000000.0,
            'err_K_prime_0': 0.2,
            'err_Debye_0': 1.0,
            'err_grueneisen_0': 0.03,
            'err_q_0': 0.2,
            'err_G_0': 2000000000.0,
            'err_Gprime_0': 0.1,
            'err_eta_s_0': 0.1}
        Mineral.__init__(self)


class fayalite (Mineral):

    def __init__(self):
        formula = 'Fe2SiO4'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Fayalite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -1370519.0,
            'V_0': 4.629e-05,
            'K_0': 1.349622e+11,
            'Kprime_0': 4.21796,
            'Debye_0': 618.7007,
            'grueneisen_0': 1.06023,
            'q_0': 3.6466,
            'G_0': 50899990000.0,
            'Gprime_0': 1.46257,
            'eta_s_0': 1.02497,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.property_modifiers = [
            ['linear', {'delta_E': 0.0, 'delta_S': 26.76, 'delta_V': 0.0}]]

        self.uncertainties = {
            'err_F_0': 1000.0,
            'err_V_0': 0.0,
            'err_K_0': 2000000000.0,
            'err_K_prime_0': 1.0,
            'err_Debye_0': 2.0,
            'err_grueneisen_0': 0.07,
            'err_q_0': 1.0,
            'err_G_0': 2000000000.0,
            'err_Gprime_0': 0.5,
            'err_eta_s_0': 0.6}
        Mineral.__init__(self)


class mg_wadsleyite (Mineral):

    def __init__(self):
        formula = 'Mg2SiO4'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Mg_Wadsleyite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -2027837.0,
            'V_0': 4.0515e-05,
            'K_0': 1.686948e+11,
            'Kprime_0': 4.3229,
            'Debye_0': 843.4973,
            'grueneisen_0': 1.2061,
            'q_0': 2.0188,
            'G_0': 1.12e+11,
            'Gprime_0': 1.44424,
            'eta_s_0': 2.63683,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 2000.0,
            'err_V_0': 0.0,
            'err_K_0': 3000000000.0,
            'err_K_prime_0': 0.2,
            'err_Debye_0': 7.0,
            'err_grueneisen_0': 0.09,
            'err_q_0': 1.0,
            'err_G_0': 2000000000.0,
            'err_Gprime_0': 0.2,
            'err_eta_s_0': 0.4}
        Mineral.__init__(self)


class fe_wadsleyite (Mineral):

    def __init__(self):
        formula = 'Fe2SiO4'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Fe_Wadsleyite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -1364668.0,
            'V_0': 4.28e-05,
            'K_0': 1.68591e+11,
            'Kprime_0': 4.3229,
            'Debye_0': 665.4492,
            'grueneisen_0': 1.2061,
            'q_0': 2.0188,
            'G_0': 72000000000.0,
            'Gprime_0': 1.44424,
            'eta_s_0': 1.04017,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.property_modifiers = [
            ['linear', {'delta_E': 0.0, 'delta_S': 26.76, 'delta_V': 0.0}]]

        self.uncertainties = {
            'err_F_0': 7000.0,
            'err_V_0': 0.0,
            'err_K_0': 13000000000.0,
            'err_K_prime_0': 1.0,
            'err_Debye_0': 21.0,
            'err_grueneisen_0': 0.3,
            'err_q_0': 1.0,
            'err_G_0': 12000000000.0,
            'err_Gprime_0': 0.5,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class mg_ringwoodite (Mineral):

    def __init__(self):
        formula = 'Mg2SiO4'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Mg_Ringwoodite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -2017557.0,
            'V_0': 3.9493e-05,
            'K_0': 1.849009e+11,
            'Kprime_0': 4.22035,
            'Debye_0': 877.7094,
            'grueneisen_0': 1.10791,
            'q_0': 2.3914,
            'G_0': 1.23e+11,
            'Gprime_0': 1.35412,
            'eta_s_0': 2.30461,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 2000.0,
            'err_V_0': 0.0,
            'err_K_0': 2000000000.0,
            'err_K_prime_0': 0.2,
            'err_Debye_0': 8.0,
            'err_grueneisen_0': 0.1,
            'err_q_0': 0.4,
            'err_G_0': 2000000000.0,
            'err_Gprime_0': 0.1,
            'err_eta_s_0': 0.5}
        Mineral.__init__(self)


class fe_ringwoodite (Mineral):

    def __init__(self):
        formula = 'Fe2SiO4'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Fe_Ringwoodite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -1362772.0,
            'V_0': 4.186e-05,
            'K_0': 2.13412e+11,
            'Kprime_0': 4.22035,
            'Debye_0': 677.7177,
            'grueneisen_0': 1.27193,
            'q_0': 2.3914,
            'G_0': 92000000000.0,
            'Gprime_0': 1.35412,
            'eta_s_0': 1.77249,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.property_modifiers = [
            ['linear', {'delta_E': 0.0, 'delta_S': 26.76, 'delta_V': 0.0}]]

        self.uncertainties = {
            'err_F_0': 2000.0,
            'err_V_0': 0.0,
            'err_K_0': 7000000000.0,
            'err_K_prime_0': 1.0,
            'err_Debye_0': 8.0,
            'err_grueneisen_0': 0.23,
            'err_q_0': 1.0,
            'err_G_0': 10000000000.0,
            'err_Gprime_0': 0.5,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class enstatite (Mineral):

    def __init__(self):
        formula = 'Mg2Si2O6'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Enstatite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -2913596.0,
            'V_0': 6.2676e-05,
            'K_0': 1.070768e+11,
            'Kprime_0': 7.02751,
            'Debye_0': 812.1848,
            'grueneisen_0': 0.78479,
            'q_0': 3.43846,
            'G_0': 76800000000.0,
            'Gprime_0': 1.54596,
            'eta_s_0': 2.50453,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 2000.0,
            'err_V_0': 0.0,
            'err_K_0': 2000000000.0,
            'err_K_prime_0': 0.4,
            'err_Debye_0': 4.0,
            'err_grueneisen_0': 0.04,
            'err_q_0': 0.4,
            'err_G_0': 1000000000.0,
            'err_Gprime_0': 0.1,
            'err_eta_s_0': 0.1}
        Mineral.__init__(self)


class ferrosilite (Mineral):

    def __init__(self):
        formula = 'Fe2Si2O6'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Ferrosilite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -2225718.0,
            'V_0': 6.5941e-05,
            'K_0': 1.005386e+11,
            'Kprime_0': 7.02751,
            'Debye_0': 674.4769,
            'grueneisen_0': 0.71889,
            'q_0': 3.43846,
            'G_0': 52000000000.0,
            'Gprime_0': 1.54596,
            'eta_s_0': 1.07706,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.property_modifiers = [
            ['linear', {'delta_E': 0.0, 'delta_S': 26.76, 'delta_V': 0.0}]]

        self.uncertainties = {
            'err_F_0': 4000.0,
            'err_V_0': 0.0,
            'err_K_0': 4000000000.0,
            'err_K_prime_0': 0.5,
            'err_Debye_0': 10.0,
            'err_grueneisen_0': 0.08,
            'err_q_0': 1.0,
            'err_G_0': 5000000000.0,
            'err_Gprime_0': 0.5,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class mg_tschermaks (Mineral):

    def __init__(self):
        formula = 'MgAl2SiO6'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Mg_Tschermaks',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -3002470.0,
            'V_0': 5.914e-05,
            'K_0': 1.070768e+11,
            'Kprime_0': 7.02751,
            'Debye_0': 783.8404,
            'grueneisen_0': 0.78479,
            'q_0': 3.43846,
            'G_0': 95950860000.0,
            'Gprime_0': 1.54596,
            'eta_s_0': 2.49099,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 9000.0,
            'err_V_0': 0.0,
            'err_K_0': 10000000000.0,
            'err_K_prime_0': 1.0,
            'err_Debye_0': 24.0,
            'err_grueneisen_0': 0.3,
            'err_q_0': 1.0,
            'err_G_0': 10000000000.0,
            'err_Gprime_0': 0.5,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class ortho_diopside (Mineral):

    def __init__(self):
        formula = 'CaMgSi2O6'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Ortho_Diopside',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -3015827.0,
            'V_0': 6.8054e-05,
            'K_0': 1.070768e+11,
            'Kprime_0': 7.02751,
            'Debye_0': 744.6988,
            'grueneisen_0': 0.78479,
            'q_0': 3.43846,
            'G_0': 58458950000.0,
            'Gprime_0': 1.54596,
            'eta_s_0': 1.36161,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 3000.0,
            'err_V_0': 0.0,
            'err_K_0': 10000000000.0,
            'err_K_prime_0': 1.0,
            'err_Debye_0': 9.0,
            'err_grueneisen_0': 0.3,
            'err_q_0': 1.0,
            'err_G_0': 10000000000.0,
            'err_Gprime_0': 0.5,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class diopside (Mineral):

    def __init__(self):
        formula = 'CaMgSi2O6'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Diopside',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -3029531.0,
            'V_0': 6.6039e-05,
            'K_0': 1.122413e+11,
            'Kprime_0': 5.23885,
            'Debye_0': 781.6146,
            'grueneisen_0': 0.95873,
            'q_0': 1.52852,
            'G_0': 67000000000.0,
            'Gprime_0': 1.37293,
            'eta_s_0': 1.57351,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 2000.0,
            'err_V_0': 0.0,
            'err_K_0': 5000000000.0,
            'err_K_prime_0': 1.8,
            'err_Debye_0': 3.0,
            'err_grueneisen_0': 0.05,
            'err_q_0': 2.0,
            'err_G_0': 2000000000.0,
            'err_Gprime_0': 0.5,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class hedenbergite (Mineral):

    def __init__(self):
        formula = 'CaFeSi2O6'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Hedenbergite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -2677330.0,
            'V_0': 6.7867e-05,
            'K_0': 1.192555e+11,
            'Kprime_0': 5.23885,
            'Debye_0': 701.5851,
            'grueneisen_0': 0.93516,
            'q_0': 1.52852,
            'G_0': 61000000000.0,
            'Gprime_0': 1.17647,
            'eta_s_0': 1.5703,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.property_modifiers = [
            ['linear', {'delta_E': 0.0, 'delta_S': 13.38, 'delta_V': 0.0}]]

        self.uncertainties = {
            'err_F_0': 45000.0,
            'err_V_0': 0.0,
            'err_K_0': 4000000000.0,
            'err_K_prime_0': 1.0,
            'err_Debye_0': 2.0,
            'err_grueneisen_0': 0.06,
            'err_q_0': 1.0,
            'err_G_0': 1000000000.0,
            'err_Gprime_0': 0.5,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class clinoenstatite (Mineral):

    def __init__(self):
        formula = 'Mg2Si2O6'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Clinoenstatite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -2905918.0,
            'V_0': 6.25e-05,
            'K_0': 1.122413e+11,
            'Kprime_0': 5.23885,
            'Debye_0': 805.0547,
            'grueneisen_0': 0.95873,
            'q_0': 1.52852,
            'G_0': 79496860000.0,
            'Gprime_0': 1.62901,
            'eta_s_0': 1.69074,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 3000.0,
            'err_V_0': 0.0,
            'err_K_0': 10000000000.0,
            'err_K_prime_0': 1.0,
            'err_Debye_0': 10.0,
            'err_grueneisen_0': 0.3,
            'err_q_0': 1.0,
            'err_G_0': 10000000000.0,
            'err_Gprime_0': 0.5,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class ca_tschermaks (Mineral):

    def __init__(self):
        formula = 'CaAl2SiO6'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Ca_Tschermaks',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -3120253.0,
            'V_0': 6.3574e-05,
            'K_0': 1.122413e+11,
            'Kprime_0': 5.23885,
            'Debye_0': 803.6626,
            'grueneisen_0': 0.78126,
            'q_0': 1.52852,
            'G_0': 75160660000.0,
            'Gprime_0': 1.54016,
            'eta_s_0': 1.9672,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.property_modifiers = [
            ['linear', {'delta_E': 0.0, 'delta_S': 11.525, 'delta_V': 0.0}]]

        self.uncertainties = {
            'err_F_0': 5000.0,
            'err_V_0': 0.0,
            'err_K_0': 10000000000.0,
            'err_K_prime_0': 1.0,
            'err_Debye_0': 5.0,
            'err_grueneisen_0': 0.0,
            'err_q_0': 1.0,
            'err_G_0': 10000000000.0,
            'err_Gprime_0': 0.5,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class jadeite (Mineral):

    def __init__(self):
        formula = 'NaAlSi2O6'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Jadeite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -2855192.0,
            'V_0': 6.0508e-05,
            'K_0': 1.422873e+11,
            'Kprime_0': 5.23885,
            'Debye_0': 820.7623,
            'grueneisen_0': 0.903,
            'q_0': 0.39234,
            'G_0': 85000000000.0,
            'Gprime_0': 1.37398,
            'eta_s_0': 2.18453,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 3000.0,
            'err_V_0': 0.0,
            'err_K_0': 2000000000.0,
            'err_K_prime_0': 1.0,
            'err_Debye_0': 12.0,
            'err_grueneisen_0': 0.08,
            'err_q_0': 1.4,
            'err_G_0': 2000000000.0,
            'err_Gprime_0': 0.5,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class hp_clinoenstatite (Mineral):

    def __init__(self):
        formula = 'Mg2Si2O6'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'HP_Clinoenstatite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -2905788.0,
            'V_0': 6.076e-05,
            'K_0': 1.160254e+11,
            'Kprime_0': 6.23685,
            'Debye_0': 824.4439,
            'grueneisen_0': 1.12473,
            'q_0': 0.20401,
            'G_0': 87927170000.0,
            'Gprime_0': 1.84119,
            'eta_s_0': 2.14181,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 3000.0,
            'err_V_0': 0.0,
            'err_K_0': 1000000000.0,
            'err_K_prime_0': 0.3,
            'err_Debye_0': 7.0,
            'err_grueneisen_0': 0.05,
            'err_q_0': 0.5,
            'err_G_0': 1000000000.0,
            'err_Gprime_0': 0.1,
            'err_eta_s_0': 0.5}
        Mineral.__init__(self)


class hp_clinoferrosilite (Mineral):

    def __init__(self):
        formula = 'Fe2Si2O6'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'HP_Clinoferrosilite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -2222183.0,
            'V_0': 6.385413e-05,
            'K_0': 1.160254e+11,
            'Kprime_0': 6.23685,
            'Debye_0': 691.564,
            'grueneisen_0': 1.12473,
            'q_0': 0.20401,
            'G_0': 70623090000.0,
            'Gprime_0': 1.84119,
            'eta_s_0': 0.79216,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.property_modifiers = [
            ['linear', {'delta_E': 0.0, 'delta_S': 26.76, 'delta_V': 0.0}]]

        self.uncertainties = {
            'err_F_0': 4000.0,
            'err_V_0': 0.0,
            'err_K_0': 10000000000.0,
            'err_K_prime_0': 1.0,
            'err_Debye_0': 11.0,
            'err_grueneisen_0': 0.3,
            'err_q_0': 1.0,
            'err_G_0': 10000000000.0,
            'err_Gprime_0': 0.5,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class ca_perovskite (Mineral):

    def __init__(self):
        formula = 'CaSiO3'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Ca_Perovskite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -1463358.0,
            'V_0': 2.745e-05,
            'K_0': 2.36e+11,
            'Kprime_0': 3.9,
            'Debye_0': 795.779,
            'grueneisen_0': 1.88839,
            'q_0': 0.89769,
            'G_0': 1.568315e+11,
            'Gprime_0': 2.22713,
            'eta_s_0': 1.28818,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 8000.0,
            'err_V_0': 0.0,
            'err_K_0': 4000000000.0,
            'err_K_prime_0': 0.2,
            'err_Debye_0': 44.0,
            'err_grueneisen_0': 0.07,
            'err_q_0': 1.6,
            'err_G_0': 12000000000.0,
            'err_Gprime_0': 0.5,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class mg_akimotoite (Mineral):

    def __init__(self):
        formula = 'MgSiO3'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Mg_Akimotoite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -1410850.0,
            'V_0': 2.6354e-05,
            'K_0': 2.10706e+11,
            'Kprime_0': 5.62088,
            'Debye_0': 935.9778,
            'grueneisen_0': 1.18984,
            'q_0': 2.34514,
            'G_0': 1.32e+11,
            'Gprime_0': 1.57889,
            'eta_s_0': 2.80782,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 2000.0,
            'err_V_0': 0.0,
            'err_K_0': 4000000000.0,
            'err_K_prime_0': 0.8,
            'err_Debye_0': 12.0,
            'err_grueneisen_0': 0.13,
            'err_q_0': 0.8,
            'err_G_0': 8000000000.0,
            'err_Gprime_0': 0.5,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class fe_akimotoite (Mineral):

    def __init__(self):
        formula = 'FeSiO3'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Fe_Akimotoite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -1067598.0,
            'V_0': 2.6854e-05,
            'K_0': 2.10706e+11,
            'Kprime_0': 5.62088,
            'Debye_0': 887.8709,
            'grueneisen_0': 1.18984,
            'q_0': 2.34514,
            'G_0': 1.523046e+11,
            'Gprime_0': 1.57889,
            'eta_s_0': 3.5716,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.property_modifiers = [
            ['linear', {'delta_E': 0.0, 'delta_S': 13.38, 'delta_V': 0.0}]]

        self.uncertainties = {
            'err_F_0': 21000.0,
            'err_V_0': 0.0,
            'err_K_0': 10000000000.0,
            'err_K_prime_0': 1.0,
            'err_Debye_0': 120.0,
            'err_grueneisen_0': 0.3,
            'err_q_0': 1.0,
            'err_G_0': 10000000000.0,
            'err_Gprime_0': 0.5,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class corundum (Mineral):

    def __init__(self):
        formula = 'AlAlO3'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Corundum',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -1582454.0,
            'V_0': 2.5577e-05,
            'K_0': 2.525457e+11,
            'Kprime_0': 4.33728,
            'Debye_0': 932.5696,
            'grueneisen_0': 1.32442,
            'q_0': 1.30316,
            'G_0': 1.632e+11,
            'Gprime_0': 1.64174,
            'eta_s_0': 2.8316,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 1000.0,
            'err_V_0': 0.0,
            'err_K_0': 5000000000.0,
            'err_K_prime_0': 0.2,
            'err_Debye_0': 3.0,
            'err_grueneisen_0': 0.04,
            'err_q_0': 0.2,
            'err_G_0': 2000000000.0,
            'err_Gprime_0': 0.1,
            'err_eta_s_0': 0.2}
        Mineral.__init__(self)


class pyrope (Mineral):

    def __init__(self):
        formula = 'Mg3Al2Si3O12'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Pyrope',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -5936538.0,
            'V_0': 0.00011308,
            'K_0': 1.702396e+11,
            'Kprime_0': 4.11067,
            'Debye_0': 823.2102,
            'grueneisen_0': 1.01424,
            'q_0': 1.42169,
            'G_0': 93699990000.0,
            'Gprime_0': 1.35756,
            'eta_s_0': 0.98186,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 10000.0,
            'err_V_0': 0.0,
            'err_K_0': 2000000000.0,
            'err_K_prime_0': 0.3,
            'err_Debye_0': 4.0,
            'err_grueneisen_0': 0.06,
            'err_q_0': 0.5,
            'err_G_0': 2000000000.0,
            'err_Gprime_0': 0.2,
            'err_eta_s_0': 0.3}
        Mineral.__init__(self)


class almandine (Mineral):

    def __init__(self):
        formula = 'Fe3Al2Si3O12'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Almandine',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -4935516.0,
            'V_0': 0.00011543,
            'K_0': 1.738963e+11,
            'Kprime_0': 4.91341,
            'Debye_0': 741.356,
            'grueneisen_0': 1.06495,
            'q_0': 1.42169,
            'G_0': 96000000000.0,
            'Gprime_0': 1.40927,
            'eta_s_0': 2.09292,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.property_modifiers = [
            ['linear', {'delta_E': 0.0, 'delta_S': 40.14, 'delta_V': 0.0}]]

        self.uncertainties = {
            'err_F_0': 29000.0,
            'err_V_0': 0.0,
            'err_K_0': 2000000000.0,
            'err_K_prime_0': 0.2,
            'err_Debye_0': 5.0,
            'err_grueneisen_0': 0.06,
            'err_q_0': 1.0,
            'err_G_0': 1000000000.0,
            'err_Gprime_0': 0.1,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class grossular (Mineral):

    def __init__(self):
        formula = 'Ca3Al2Si3O12'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Grossular',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -6277935.0,
            'V_0': 0.00012512,
            'K_0': 1.670622e+11,
            'Kprime_0': 3.91544,
            'Debye_0': 822.743,
            'grueneisen_0': 1.05404,
            'q_0': 1.88887,
            'G_0': 1.09e+11,
            'Gprime_0': 1.16274,
            'eta_s_0': 2.38418,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 11000.0,
            'err_V_0': 0.0,
            'err_K_0': 1000000000.0,
            'err_K_prime_0': 0.2,
            'err_Debye_0': 2.0,
            'err_grueneisen_0': 0.06,
            'err_q_0': 0.2,
            'err_G_0': 4000000000.0,
            'err_Gprime_0': 0.1,
            'err_eta_s_0': 0.1}
        Mineral.__init__(self)


class mg_majorite (Mineral):

    def __init__(self):
        formula = 'Mg4Si4O12'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Mg_Majorite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -5691614.0,
            'V_0': 0.000114324,
            'K_0': 1.651183e+11,
            'Kprime_0': 4.21183,
            'Debye_0': 822.458,
            'grueneisen_0': 0.97682,
            'q_0': 1.53581,
            'G_0': 84999990000.0,
            'Gprime_0': 1.42969,
            'eta_s_0': 1.0178,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 10000.0,
            'err_V_0': 0.0,
            'err_K_0': 3000000000.0,
            'err_K_prime_0': 0.3,
            'err_Debye_0': 4.0,
            'err_grueneisen_0': 0.07,
            'err_q_0': 0.5,
            'err_G_0': 2000000000.0,
            'err_Gprime_0': 0.2,
            'err_eta_s_0': 0.3}
        Mineral.__init__(self)


class jd_majorite (Mineral):

    def __init__(self):
        formula = 'Na2Al2Si4O12'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Jd_Majorite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -5518542.0,
            'V_0': 0.00011094,
            'K_0': 1.770772e+11,
            'Kprime_0': 4.11067,
            'Debye_0': 895.914,
            'grueneisen_0': 1.01424,
            'q_0': 1.42169,
            'G_0': 1.25e+11,
            'Gprime_0': 1.35756,
            'eta_s_0': 3.30517,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 14000.0,
            'err_V_0': 0.0,
            'err_K_0': 7000000000.0,
            'err_K_prime_0': 1.0,
            'err_Debye_0': 18.0,
            'err_grueneisen_0': 0.3,
            'err_q_0': 1.0,
            'err_G_0': 4000000000.0,
            'err_Gprime_0': 0.5,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class quartz (Mineral):

    def __init__(self):
        formula = 'SiO2'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Quartz',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -858853.4,
            'V_0': 2.367003e-05,
            'K_0': 49547430000.0,
            'Kprime_0': 4.33155,
            'Debye_0': 816.3307,
            'grueneisen_0': -0.00296,
            'q_0': 1.0,
            'G_0': 44856170000.0,
            'Gprime_0': 0.95315,
            'eta_s_0': 2.36469,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.property_modifiers = [
            ['landau', {'Tc_0': 847.0, 'S_D': 5.164, 'V_D': 1.222e-06}]]

        self.uncertainties = {
            'err_F_0': 1000.0,
            'err_V_0': 0.0,
            'err_K_0': 1000000000.0,
            'err_K_prime_0': 0.1,
            'err_Debye_0': 31.0,
            'err_grueneisen_0': 0.05,
            'err_q_0': 1.0,
            'err_G_0': 1000000000.0,
            'err_Gprime_0': 0.1,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class coesite (Mineral):

    def __init__(self):
        formula = 'SiO2'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Coesite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -855068.5,
            'V_0': 2.0657e-05,
            'K_0': 1.135856e+11,
            'Kprime_0': 4.0,
            'Debye_0': 852.4267,
            'grueneisen_0': 0.39157,
            'q_0': 1.0,
            'G_0': 61600010000.0,
            'Gprime_0': 1.24734,
            'eta_s_0': 2.39793,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 1000.0,
            'err_V_0': 0.0,
            'err_K_0': 1000000000.0,
            'err_K_prime_0': 1.0,
            'err_Debye_0': 9.0,
            'err_grueneisen_0': 0.05,
            'err_q_0': 1.0,
            'err_G_0': 1000000000.0,
            'err_Gprime_0': 0.5,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class stishovite (Mineral):

    def __init__(self):
        formula = 'SiO2'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Stishovite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -818984.6,
            'V_0': 1.4017e-05,
            'K_0': 3.143352e+11,
            'Kprime_0': 3.75122,
            'Debye_0': 1107.824,
            'grueneisen_0': 1.37466,
            'q_0': 2.83517,
            'G_0': 2.2e+11,
            'Gprime_0': 1.93334,
            'eta_s_0': 4.60904,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.property_modifiers = [
            ['landau', {'Tc_0': -4250.0, 'S_D': 0.012, 'V_D': 1e-09}]]

        self.uncertainties = {
            'err_F_0': 1000.0,
            'err_V_0': 0.0,
            'err_K_0': 8000000000.0,
            'err_K_prime_0': 0.1,
            'err_Debye_0': 13.0,
            'err_grueneisen_0': 0.17,
            'err_q_0': 2.2,
            'err_G_0': 12000000000.0,
            'err_Gprime_0': 0.1,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class seifertite (Mineral):

    def __init__(self):
        formula = 'SiO2'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Seifertite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -794335.4,
            'V_0': 1.367e-05,
            'K_0': 3.275843e+11,
            'Kprime_0': 4.01553,
            'Debye_0': 1140.772,
            'grueneisen_0': 1.37466,
            'q_0': 2.83517,
            'G_0': 2.274532e+11,
            'Gprime_0': 1.76965,
            'eta_s_0': 4.97108,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 2000.0,
            'err_V_0': 0.0,
            'err_K_0': 2000000000.0,
            'err_K_prime_0': 0.1,
            'err_Debye_0': 16.0,
            'err_grueneisen_0': 0.3,
            'err_q_0': 1.0,
            'err_G_0': 2000000000.0,
            'err_Gprime_0': 0.1,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class mg_perovskite (Mineral):

    def __init__(self):
        formula = 'MgSiO3'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Mg_Perovskite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -1368283.0,
            'V_0': 2.4445e-05,
            'K_0': 2.505264e+11,
            'Kprime_0': 4.14,
            'Debye_0': 905.9412,
            'grueneisen_0': 1.56508,
            'q_0': 1.10945,
            'G_0': 1.729e+11,
            'Gprime_0': 1.69037,
            'eta_s_0': 2.56536,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 1000.0,
            'err_V_0': 0.0,
            'err_K_0': 3000000000.0,
            'err_K_prime_0': 0.1,
            'err_Debye_0': 5.0,
            'err_grueneisen_0': 0.05,
            'err_q_0': 0.3,
            'err_G_0': 2000000000.0,
            'err_Gprime_0': 0.0,
            'err_eta_s_0': 0.3}
        Mineral.__init__(self)


class fe_perovskite (Mineral):

    def __init__(self):
        formula = 'FeSiO3'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Fe_Perovskite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -1040920.0,
            'V_0': 2.5485e-05,
            'K_0': 2.721152e+11,
            'Kprime_0': 4.14,
            'Debye_0': 870.8122,
            'grueneisen_0': 1.56508,
            'q_0': 1.10945,
            'G_0': 1.326849e+11,
            'Gprime_0': 1.37485,
            'eta_s_0': 2.29211,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.property_modifiers = [
            ['linear', {'delta_E': 0.0, 'delta_S': 13.38, 'delta_V': 0.0}]]

        self.uncertainties = {
            'err_F_0': 6000.0,
            'err_V_0': 0.0,
            'err_K_0': 40000000000.0,
            'err_K_prime_0': 1.0,
            'err_Debye_0': 26.0,
            'err_grueneisen_0': 0.3,
            'err_q_0': 1.0,
            'err_G_0': 40000000000.0,
            'err_Gprime_0': 0.0,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class al_perovskite (Mineral):

    def __init__(self):
        formula = 'AlAlO3'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Al_perovskite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -1533878.0,
            'V_0': 2.4944e-05,
            'K_0': 2.582e+11,
            'Kprime_0': 4.14,
            'Debye_0': 886.4601,
            'grueneisen_0': 1.56508,
            'q_0': 1.10945,
            'G_0': 1.713116e+11,
            'Gprime_0': 1.49706,
            'eta_s_0': 2.47126,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 2000.0,
            'err_V_0': 0.0,
            'err_K_0': 10000000000.0,
            'err_K_prime_0': 0.5,
            'err_Debye_0': 7.0,
            'err_grueneisen_0': 0.3,
            'err_q_0': 1.0,
            'err_G_0': 10000000000.0,
            'err_Gprime_0': 0.1,
            'err_eta_s_0': 0.5}
        Mineral.__init__(self)


class mg_post_perovskite (Mineral):

    def __init__(self):
        formula = 'MgSiO3'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Mg_Post_Perovskite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -1348641.0,
            'V_0': 2.4419e-05,
            'K_0': 2.312e+11,
            'Kprime_0': 4.0,
            'Debye_0': 855.8173,
            'grueneisen_0': 1.89155,
            'q_0': 1.09081,
            'G_0': 1.50167e+11,
            'Gprime_0': 1.97874,
            'eta_s_0': 1.16704,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 3000.0,
            'err_V_0': 0.0,
            'err_K_0': 1000000000.0,
            'err_K_prime_0': 0.1,
            'err_Debye_0': 7.0,
            'err_grueneisen_0': 0.03,
            'err_q_0': 0.1,
            'err_G_0': 4000000000.0,
            'err_Gprime_0': 0.1,
            'err_eta_s_0': 0.2}
        Mineral.__init__(self)


class fe_post_perovskite (Mineral):

    def __init__(self):
        formula = 'FeSiO3'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Fe_Post_Perovskite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -981806.9,
            'V_0': 2.5459e-05,
            'K_0': 2.312e+11,
            'Kprime_0': 4.0,
            'Debye_0': 781.3465,
            'grueneisen_0': 1.89155,
            'q_0': 1.09081,
            'G_0': 1.295e+11,
            'Gprime_0': 1.44675,
            'eta_s_0': 1.36382,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.property_modifiers = [
            ['linear', {'delta_E': 0.0, 'delta_S': 13.38, 'delta_V': 0.0}]]

        self.uncertainties = {
            'err_F_0': 21000.0,
            'err_V_0': 0.0,
            'err_K_0': 10000000000.0,
            'err_K_prime_0': 1.0,
            'err_Debye_0': 52.0,
            'err_grueneisen_0': 0.3,
            'err_q_0': 1.0,
            'err_G_0': 5000000000.0,
            'err_Gprime_0': 0.1,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class al_post_perovskite (Mineral):

    def __init__(self):
        formula = 'AlAlO3'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Al_Post_Perovskite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -1377582.0,
            'V_0': 2.3847e-05,
            'K_0': 2.49e+11,
            'Kprime_0': 4.0,
            'Debye_0': 762.1951,
            'grueneisen_0': 1.64573,
            'q_0': 1.09081,
            'G_0': 91965310000.0,
            'Gprime_0': 1.81603,
            'eta_s_0': 2.83762,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 4000.0,
            'err_V_0': 0.0,
            'err_K_0': 20000000000.0,
            'err_K_prime_0': 0.1,
            'err_Debye_0': 9.0,
            'err_grueneisen_0': 0.02,
            'err_q_0': 1.0,
            'err_G_0': 10000000000.0,
            'err_Gprime_0': 0.1,
            'err_eta_s_0': 0.2}
        Mineral.__init__(self)


class periclase (Mineral):

    def __init__(self):
        formula = 'MgO'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Periclase',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -569444.6,
            'V_0': 1.1244e-05,
            'K_0': 1.613836e+11,
            'Kprime_0': 3.84045,
            'Debye_0': 767.0977,
            'grueneisen_0': 1.36127,
            'q_0': 1.7217,
            'G_0': 1.309e+11,
            'Gprime_0': 2.1438,
            'eta_s_0': 2.81765,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 0.0,
            'err_V_0': 0.0,
            'err_K_0': 3000000000.0,
            'err_K_prime_0': 0.2,
            'err_Debye_0': 9.0,
            'err_grueneisen_0': 0.05,
            'err_q_0': 0.2,
            'err_G_0': 1000000000.0,
            'err_Gprime_0': 0.1,
            'err_eta_s_0': 0.2}
        Mineral.__init__(self)


class wuestite (Mineral):

    def __init__(self):
        formula = 'FeO'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Wuestite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -242146.0,
            'V_0': 1.2264e-05,
            'K_0': 1.794442e+11,
            'Kprime_0': 4.9376,
            'Debye_0': 454.1592,
            'grueneisen_0': 1.53047,
            'q_0': 1.7217,
            'G_0': 59000000000.0,
            'Gprime_0': 1.44673,
            'eta_s_0': -0.05731,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.property_modifiers = [
            ['linear', {'delta_E': 0.0, 'delta_S': 13.38, 'delta_V': 0.0}]]

        self.uncertainties = {
            'err_F_0': 1000.0,
            'err_V_0': 0.0,
            'err_K_0': 1000000000.0,
            'err_K_prime_0': 0.2,
            'err_Debye_0': 21.0,
            'err_grueneisen_0': 0.13,
            'err_q_0': 1.0,
            'err_G_0': 1000000000.0,
            'err_Gprime_0': 0.1,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class mg_ca_ferrite (Mineral):

    def __init__(self):
        formula = 'MgAl2O4'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Mg_Ca_Ferrite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -2122169.0,
            'V_0': 3.6177e-05,
            'K_0': 2.106663e+11,
            'Kprime_0': 4.0528,
            'Debye_0': 838.6291,
            'grueneisen_0': 1.31156,
            'q_0': 1.0,
            'G_0': 1.29826e+11,
            'Gprime_0': 1.75878,
            'eta_s_0': 2.1073,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 4000.0,
            'err_V_0': 0.0,
            'err_K_0': 1000000000.0,
            'err_K_prime_0': 0.1,
            'err_Debye_0': 16.0,
            'err_grueneisen_0': 0.3,
            'err_q_0': 1.0,
            'err_G_0': 1000000000.0,
            'err_Gprime_0': 0.1,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class fe_ca_ferrite (Mineral):

    def __init__(self):
        formula = 'FeAl2O4'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Fe_Ca_Ferrite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -1790284.0,
            'V_0': 3.7258e-05,
            'K_0': 2.106663e+11,
            'Kprime_0': 4.0528,
            'Debye_0': 804.1986,
            'grueneisen_0': 1.31156,
            'q_0': 1.0,
            'G_0': 1.535236e+11,
            'Gprime_0': 1.75878,
            'eta_s_0': 3.0268,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.property_modifiers = [
            ['linear', {'delta_E': 0.0, 'delta_S': 13.38, 'delta_V': 0.0}]]

        self.uncertainties = {
            'err_F_0': 25000.0,
            'err_V_0': 0.0,
            'err_K_0': 10000000000.0,
            'err_K_prime_0': 1.0,
            'err_Debye_0': 69.0,
            'err_grueneisen_0': 0.3,
            'err_q_0': 1.0,
            'err_G_0': 10000000000.0,
            'err_Gprime_0': 0.5,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class na_ca_ferrite (Mineral):

    def __init__(self):
        formula = 'NaAlSiO4'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Na_Ca_Ferrite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -1844129.0,
            'V_0': 3.627e-05,
            'K_0': 1.613385e+11,
            'Kprime_0': 4.32479,
            'Debye_0': 812.4769,
            'grueneisen_0': 0.69428,
            'q_0': 1.0,
            'G_0': 1.220049e+11,
            'Gprime_0': 2.07687,
            'eta_s_0': 2.79016,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 11000.0,
            'err_V_0': 0.0,
            'err_K_0': 1000000000.0,
            'err_K_prime_0': 0.1,
            'err_Debye_0': 51.0,
            'err_grueneisen_0': 0.3,
            'err_q_0': 1.0,
            'err_G_0': 1000000000.0,
            'err_Gprime_0': 0.1,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class kyanite (Mineral):

    def __init__(self):
        formula = 'Al2SiO5'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Kyanite',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -2446058.0,
            'V_0': 4.4227e-05,
            'K_0': 1.6e+11,
            'Kprime_0': 4.0,
            'Debye_0': 943.1665,
            'grueneisen_0': 0.9255,
            'q_0': 1.0,
            'G_0': 1.204033e+11,
            'Gprime_0': 1.7308,
            'eta_s_0': 2.96665,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 4000.0,
            'err_V_0': 0.0,
            'err_K_0': 1000000000.0,
            'err_K_prime_0': 0.0,
            'err_Debye_0': 8.0,
            'err_grueneisen_0': 0.07,
            'err_q_0': 1.0,
            'err_G_0': 10000000000.0,
            'err_Gprime_0': 0.5,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)


class nepheline (Mineral):

    def __init__(self):
        formula = 'NaAlSiO4'
        formula = dictionarize_formula(formula)
        self.params = {
            'name': 'Nepheline',
            'formula': formula,
            'equation_of_state': 'slb3',
            'F_0': -1992104.0,
            'V_0': 5.46684e-05,
            'K_0': 53077990000.0,
            'Kprime_0': 4.0,
            'Debye_0': 700.9422,
            'grueneisen_0': 0.69428,
            'q_0': 1.0,
            'G_0': 30700000000.0,
            'Gprime_0': 1.33031,
            'eta_s_0': 0.6291,
            'n': sum(formula.values()),
            'molar_mass': formula_mass(formula)}

        self.uncertainties = {
            'err_F_0': 3000.0,
            'err_V_0': 0.0,
            'err_K_0': 1000000000.0,
            'err_K_prime_0': 1.0,
            'err_Debye_0': 13.0,
            'err_grueneisen_0': 0.03,
            'err_q_0': 1.0,
            'err_G_0': 1000000000.0,
            'err_Gprime_0': 0.5,
            'err_eta_s_0': 1.0}
        Mineral.__init__(self)

'''
Mineral aliases
'''

# Feldspars
ab = albite
an = anorthite

# LP Spinels
sp = spinel
hc = hercynite

# Olivine polymorphs
fo = forsterite
fa = fayalite
mgwa = mg_wadsleyite
fewa = fe_wadsleyite
mgri = mg_ringwoodite
feri = fe_ringwoodite

# Orthopyroxenes
en = enstatite
fs = ferrosilite
mgts = mg_tschermaks
odi = ortho_diopside

# Clinopyroxenes
di = diopside
he = hedenbergite
cen = clinoenstatite
cats = ca_tschermaks
jd = jadeite
mgc2 = hp_clinoenstatite
fec2 = hp_clinoferrosilite
hpcen = hp_clinoenstatite
hpcfs = hp_clinoferrosilite

# Perovskites
mgpv = mg_perovskite
mg_bridgmanite = mg_perovskite
fepv = fe_perovskite
fe_bridgmanite = fe_perovskite
alpv = al_perovskite
capv = ca_perovskite

# Ilmenite group
mgil = mg_akimotoite
feil = fe_akimotoite
co = corundum

# Garnet group
py = pyrope
al = almandine
gr = grossular
mgmj = mg_majorite
jdmj = jd_majorite

# Quartz polymorphs
qtz = quartz
coes = coesite
st = stishovite
seif = seifertite

# Post perovskites
mppv = mg_post_perovskite
fppv = fe_post_perovskite
appv = al_post_perovskite

# Magnesiowuestite
pe = periclase
wu = wuestite

# Calcium ferrite structured phases
mgcf = mg_ca_ferrite
fecf = fe_ca_ferrite
nacf = na_ca_ferrite

# Al2SiO5 polymorphs
ky = kyanite

# Nepheline group
neph = nepheline


# Solid solution aliases
c2c = c2c_pyroxene
cf = ca_ferrite_structured_phase
cpx = clinopyroxene
gt = garnet
il = akimotoite
ilmenite_group = akimotoite
mw = ferropericlase
magnesiowuestite = ferropericlase
ol = mg_fe_olivine
opx = orthopyroxene
plag = plagioclase
ppv = post_perovskite
pv = mg_fe_perovskite
mg_fe_bridgmanite = mg_fe_perovskite
mg_fe_silicate_perovskite = mg_fe_perovskite
ri = mg_fe_ringwoodite
spinel_group = mg_fe_aluminous_spinel
wa = mg_fe_wadsleyite
spinelloid_III = mg_fe_wadsleyite



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.


# This is a standalone program that converts the Holland and Powell data format
# into the standard burnman format (printed to file)
# It only outputs properties of solid and melt endmembers
# - other endmembers are currently ignored.


import sys
import os.path
import pprint
from collections import OrderedDict
# hack to allow scripts to be placed in subdirectories next to burnman:
if not os.path.exists('burnman') and os.path.exists('../../../burnman'):
    sys.path.insert(1, os.path.abspath('../../..'))

from burnman.tools.chemistry import dictionarize_formula, formula_mass


if not os.path.isfile('tc-ds633.txt'):
    print('This code requires the data file tc-ds633.txt.')
    print(
        'This file is bundled with the software package THERMOCALC, which can be found here:')
    print(
        'https://www.esc.cam.ac.uk/research/research-groups/research-projects/tim-hollands-software-pages/thermocalc')
    print('')
    print('Please download the file and place it in this directory.')
    exit()

# Components
components = ['Si', 'Ti', 'Al', 'Fe', 'Mg', 'Mn', 'Ca', 'Na',
              'K', 'O', 'H', 'C', 'Cl', 'e-', 'Ni', 'Zr', 'S', 'Cu', 'Cr']


class Endmember:

    def __init__(self, name, atoms, formula, sites, comp,
                 H, S, V, Cp, a, k, flag, od):
        if flag != -1 and flag != -2 and k[0] > 0:
            formula = dictionarize_formula(formula)
            self.params = OrderedDict([('name', name),
                                       ('formula', formula),
                                       ('equation_of_state', 'hp_tmt'),
                                       ('H_0', round(H * 1e3, 10)),
                                       ('S_0', round(S * 1e3, 10)),
                                       ('V_0', round(V * 1e-5, 15)),
                                       ('Cp', [round(Cp[0] * 1e3, 10),
                                               round(Cp[1] * 1e3, 10),
                                               round(Cp[2] * 1e3, 10),
                                               round(Cp[3] * 1e3, 10)]),
                                       ('a_0', a),
                                       ('K_0', round(k[0] * 1e8, 10)),
                                       ('Kprime_0', k[1]),
                                       ('Kdprime_0', round(k[2] * 1e-8, 15)),
                                       ('n', sum(formula.values())),
                                       ('molar_mass',
                                        round(formula_mass(formula), 10))])
        if flag == 1:
            self.landau_hp = OrderedDict([('P_0', 1e5),
                                          ('T_0', 298.15),
                                          ('Tc_0', od[0]),
                                          ('S_D', round(od[1] * 1e3, 10)),
                                          ('V_D', round(od[2] * 1e-5, 10))])

        if flag == 2:
            self.bragg_williams = OrderedDict([('deltaH',
                                                round(od[0] * 1e3, 10)),
                                               ('deltaV',
                                                round(od[1] * 1e-5, 15)),
                                               ('Wh', round(od[2] * 1e3, 10)),
                                               ('Wv', round(od[3] * 1e-5, 15)),
                                               ('n', od[4]),
                                               ('factor', od[5])])

        if flag == -2:
            formula = dictionarize_formula(formula)
            self.params = OrderedDict([('name', name),
                                       ('formula', formula),
                                       ('equation_of_state', 'hp_tmtL'),
                                       ('H_0', round(H * 1e3, 10)),
                                       ('S_0', round(S * 1e3, 10)),
                                       ('V_0', round(V * 1e-5, 15)),
                                       ('Cp', [round(Cp[0] * 1e3, 10),
                                               round(Cp[1] * 1e3, 10),
                                               round(Cp[2] * 1e3, 10),
                                               round(Cp[3] * 1e3, 10)]),
                                       ('a_0', a),
                                       ('K_0', round(k[0] * 1e8, 10)),
                                       ('Kprime_0', k[1]),
                                       ('Kdprime_0', round(k[2] * 1e-8, 15)),
                                       ('dKdT_0', round(k[3] * 1e8, 15)),
                                       ('n', sum(formula.values())),
                                       ('molar_mass',
                                        round(formula_mass(formula), 10))])


# Read dataset
with open('tc-ds633.txt', 'r') as f:
    ds = [line.split() for line in f]


def getmbr(ds, mbr):
    for i in range(0, int(ds[0][0])):
        if ds[i * 4 + 3][0] == mbr:
            atoms = 0.0
            formula = ''
            for j in range(3, len(ds[i * 4 + 3]) - 1, 2):
                atoms += float(ds[i * 4 + 3][j])
                formula = formula + \
                    components[int(ds[i * 4 + 3][j - 1]) - 1] + str(
                        round(float(ds[i * 4 + 3][j]), 10))

            sites = int(ds[i * 4 + 3][1])
            comp = list(map(float, ds[i * 4 + 3][2:(len(ds[i * 4 + 3]) - 1)]))
            H = float(ds[i * 4 + 4][0])
            S = float(ds[i * 4 + 4][1])
            V = float(ds[i * 4 + 4][2])
            Cp = list(map(float, ds[i * 4 + 5]))
            a = float(ds[i * 4 + 6][0])
            k = list(map(float, ds[i * 4 + 6][1:4]))
            od = list(map(float, ds[i * 4 + 6][5:]))

            if mbr.endswith('L'):
                flag = -2
                od = [0]
                k.append(float(ds[i * 4 + 6][4]))
            else:
                flag = int(ds[i * 4 + 6][4])

            endmember = Endmember(mbr, atoms, formula, sites,
                                  comp, H, S, V, Cp, a, k, flag, od)
            return endmember


with open('HGP_2018_ds633.py', 'w') as outfile:
    outfile.write('# This file is part of BurnMan - a thermoelastic and '
                  'thermodynamic toolkit for the Earth and Planetary '
                  'Sciences\n'
                  '# Copyright (C) 2012 - 2021 by the BurnMan team, '
                  'released under the GNU\n# GPL v2 or later.\n\n\n'
                  '"""\n'
                  'HGP_2018_ds633\n'
                  '^^^^^^^^^^^^^^\n'
                  '\n'
                  'Endmember minerals from Holland, Green and Powell (2018) and references therein.\n'
                  'Dataset version 6.33.\n'
                  'The values in this document are all in S.I. units,\n'
                  'unlike those in the original tc-ds633.txt.\n'
                  'File autogenerated using HGP633data_to_burnman.py.\n'
                  '"""\n\n'
                  'from ..mineral import Mineral\n\n')

    outfile.write('"""\n'
                  'ENDMEMBERS\n'
                  '"""\n\n')

    def pprint_ordered_dict(d, leading_string, extra_whitespace=0, width=200):
        old_leading_string = 'OrderedDict['
        whitespace = ' ' * (len(leading_string) + extra_whitespace
                            - len(old_leading_string))
        s = pprint.pformat(d, width=width)
        s = s.replace('),\n', f',\n{whitespace}')
        s = s.replace('), ', f',\n{whitespace}')
        s = s.replace('\', ', '\': ').replace('(', '')
        s = s.replace(old_leading_string,
                      leading_string+'{').replace(')])', '}')
        return s

    formula = '0'
    for i in range(int(ds[0][0])):
        mbr = ds[i * 4 + 3][0]
        M = getmbr(ds, mbr)
        if mbr == 'and':  # change silly abbreviation
            mbr = 'andalusite'

        # Print parameters
        if hasattr(M, 'params'):
            outfile.write('\n')
            outfile.write('class {0} (Mineral):\n'.format(mbr)
                          + '    def __init__(self):\n')

            s = pprint_ordered_dict(
                M.params, leading_string='        self.params = ')
            s = s.replace('000000.0', 'e6')
            outfile.write(s)
            outfile.write('\n')

            # Print property modifiers (if they exist)
            if hasattr(M, 'landau_hp'):
                outfile.write('        self.property_modifiers = [[')
                s = pprint_ordered_dict(M.landau_hp,
                                        leading_string='\'landau_hp\', ',
                                        extra_whitespace=49)
                outfile.write(s)
                outfile.write(']]\n')

            if hasattr(M, 'bragg_williams'):
                outfile.write('        self.property_modifiers = [[')
                s = pprint_ordered_dict(M.bragg_williams,
                                        leading_string='\'bragg_williams\', ',
                                        extra_whitespace=49)
                outfile.write(s)
                outfile.write(']]\n')

            outfile.write('        Mineral.__init__(self)\n\n')

    outfile.write('\ndef cov():\n'
                  '    \"\"\"\n'
                  '    A function which loads and returns the '
                  'variance-covariance matrix of the\n'
                  '    zero-point energies of all the endmembers in the '
                  'dataset.\n\n'
                  '    Returns\n'
                  '    -------\n'
                  '    cov : dictionary\n'
                  '        Dictionary keys are:\n'
                  '        - endmember_names: a list of endmember names, and\n'
                  '        - covariance_matrix: a 2D variance-covariance '
                  'array for the\n'
                  '          endmember zero-point energies of formation\n'
                  '    \"\"\"\n\n'
                  '    from .HGP_2018_ds633_cov import cov\n'
                  '    return cov\n')

# Process uncertainties
with open('HGP_2018_ds633_cov.py', 'w') as outfile:

    outfile.write('# This file is part of BurnMan - a thermoelastic and '
                  'thermodynamic toolkit for the Earth and Planetary '
                  'Sciences\n'
                  '# Copyright (C) 2012 - 2021 by the BurnMan team, '
                  'released under the GNU\n# GPL v2 or later.\n\n\n'
                  '"""\n'
                  'HGP_2018 (ds-633) zero-point energy covariance matrix\n'
                  'Derived from Holland, Green and Powell (2018) and '
                  'references therein.\n'
                  'Dataset version 6.33.\n'
                  'The values in this document are all in S.I. units,\n'
                  'unlike those in the original tc-ds633.txt.\n'
                  'File autogenerated using HGP633data_to_burnman.py.\n'
                  '"""\n\n'
                  'from numpy import array\n\n'
                  'cov = ')

    import numpy as np
    n_mbrs = int(ds[0][0])

    names = []
    for i in range(n_mbrs):
        names.append(ds[i*4+3][0])

    cov = []
    for i in range(n_mbrs*4+4, len(ds)-2):
        cov.extend(map(float, ds[i]))

    i_utr = np.triu_indices(n_mbrs)
    i_ltr = np.tril_indices(n_mbrs)
    M = np.zeros((n_mbrs, n_mbrs))

    M[i_utr] = cov[1:]
    M[i_ltr] = M.T[i_ltr]

    M = M*1.e6  # (kJ/mol)^2 -> (J/mol)^2

    d = {'endmember_names': names,
         'covariance_matrix': M}

    np.set_printoptions(threshold=sys.maxsize)

    pp = pprint.PrettyPrinter(indent=0, width=160, depth=3, stream=outfile)
    pp.pprint(d)

    outfile.write('\n')



# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for
# the Earth and Planetary Sciences
# Copyright (C) 2012 - 2021 by the BurnMan team, released under the GNU
# GPL v2 or later.

"""
BurnMan object classes
"""

from __future__ import absolute_import

from . import material
from . import perplex
from . import mineral
from . import combinedmineral
from . import solutionmodel
from . import solidsolution
from . import composite
from . import anisotropy
from . import anisotropicmineral
from . import mineral_helpers



from __future__ import absolute_import
# This file is part of BurnMan - a thermoelastic and thermodynamic toolkit for the Earth and Planetary Sciences
# Copyright (C) 2012 - 2017 by the BurnMan team, released under the GNU
# GPL v2 or later.

import numpy as np
import scipy.optimize as opt
from . import equation_of_state as eos
from ..tools.math import bracket
import warnings


def bulk_modulus_fourth(volume, params):
    """
    compute the bulk modulus as per the fourth order
    birch-murnaghan equation of state.  Returns bulk
    modulus in the same units as the reference bulk
    modulus.  Pressure must be in :math:`[Pa]`.
    """

    x = params['V_0'] / volume
    f = 0.5 * (pow(x, 2. / 3.) - 1.0)

    Xi = (3. / 4.) * (4. - params['Kprime_0'])
    Zeta = (3. / 8.) * ((params['K_0'] * params['Kprime_prime_0']) + params[
        'Kprime_0'] * (params['Kprime_0'] - 7.) + 143. / 9.)

    K = (5. * f * pow((1. + 2. * f), 5. / 2.) * params['K_0'] * (1. - (2. * Xi * f) + (4. * Zeta * pow(f, 2.)))) + \
        (pow(1. + (2. * f), 7. / 2.) * params['K_0'] * (
            1. - (4. * Xi * f) + (12. * Zeta * pow(f, 2.))))

    return K


def volume_fourth_order(pressure, params):
    func = lambda x: birch_murnaghan_fourth(
        params['V_0'] / x, params) - pressure
    try:
        sol = bracket(func, params['V_0'], 1.e-2 * params['V_0'])
    except:
        raise ValueError(
            'Cannot find a volume, perhaps you are outside of the range of validity for the equation of state?')
    return opt.brentq(func, sol[0], sol[1])


def birch_murnaghan_fourth(x, params):
    """
    equation for the fourth order birch-murnaghan equation of state, returns
    pressure in the same units that are supplied for the reference bulk
    modulus (params['K_0'])
    """

    f = 0.5 * (pow(x, 2. / 3.) - 1.0)
    Xi = (3. / 4.) * (4. - params['Kprime_0'])
    Zeta = (3. / 8.) * ((params['K_0'] * params['Kprime_prime_0']) + params[
        'Kprime_0'] * (params['Kprime_0'] - 7.) + 143. / 9.)

    return 3. * f * pow(1. + 2. * f, 5. / 2.) * params['K_0'] * (1. - (2. * Xi * f) + (4. * Zeta * pow(f, 2.))) + params['P_0']


class BM4(eos.EquationOfState):

    """
    Base class for the isothermal Birch Murnaghan equation of state.  This is fourth order in strain, and
    has no temperature dependence.
    """

    def volume(self, pressure, temperature, params):
        """
        Returns volume :math:`[m^3]` as a function of pressure :math:`[Pa]`.
        """
        return volume_fourth_order(pressure, params)

    def pressure(self, temperature, volume, params):
        return birch_murnaghan_fourth(volume / params['V_0'], params)

    def isothermal_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns isothermal bulk modulus :math:`K_T` :math:`[Pa]` as a function of pressure :math:`[Pa]`,
        temperature :math:`[K]` and volume :math:`[m^3]`.
        """
        return bulk_modulus_fourth(volume, params)

    def adiabatic_bulk_modulus(self, pressure, temperature, volume, params):
        """
        Returns adiabatic bulk modulus :math:`K_s` of the mineral. :math:`[Pa]`.
        """
        return bulk_modulus_fourth(volume, params)

    def shear_modulus(self, pressure, temperature, volume, params):
        """
        Returns shear modulus :math:`G` of the mineral. :math:`[Pa]`
        """
        return 0.

    def entropy(self, pressure, temperature, volume, params):
        """
        Returns the molar entropy :math:`\mathcal{S}` of the mineral. :math:`[J/K/mol]`
        """
        return 0.

    def molar_internal_energy(self, pressure, temperature, volume, params):
        """
        Returns the internal energy :math:`\mathcal{E}` of the mineral. :math:`[J/mol]`
        """
        x = np.power(volume/params['V_0'], -1./3.)
        x2 = x*x
        x4 = x2*x2
        x6 = x4*x2
        x8 = x4*x4

        xi1 = 3.*(4. - params['Kprime_0'])/4.
        xi2 = 3./8.*(params['K_0'] *
                     params['Kprime_prime_0'] +
                     params['Kprime_0'] *
                     (params['Kprime_0'] - 7.)) + 143./24.

        intPdV = (-9./2. * params['V_0'] * params['K_0'] *
                  ((xi1 + 1.)*(x4/4. - x2/2. + 1./4.) -
                   xi1*(x6/6. - x4/4. + 1./12.) +
                   xi2*(x8/8 - x6/2 + 3.*x4/4. - x2/2. + 1./8.)))

        return - intPdV + params['E_0']

    def gibbs_free_energy(self, pressure, temperature, volume, params):
        """
        Returns the Gibbs free energy :math:`\mathcal{G}` of the mineral. :math:`[J/mol]`
        """
        # G = int VdP = [PV] - int PdV = E + PV

        return self.molar_internal_energy(pressure, temperature, volume, params) + volume*pressure

    def molar_heat_capacity_v(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return a very large number. :math:`[J/K/mol]`
        """
        return 1.e99

    def molar_heat_capacity_p(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return a very large number. :math:`[J/K/mol]`
        """
        return 1.e99

    def thermal_expansivity(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return zero. :math:`[1/K]`
        """
        return 0.

    def grueneisen_parameter(self, pressure, temperature, volume, params):
        """
        Since this equation of state does not contain temperature effects, simply return zero. :math:`[unitless]`
        """
        return 0.

    def validate_parameters(self, params):
        """
        Check for existence and validity of the parameters
        """

        if 'E_0' not in params:
            params['E_0'] = 0.
        if 'P_0' not in params:
            params['P_0'] = 0.

        # If G and Gprime are not included this is presumably deliberate,
        # as we can model density and bulk modulus just fine without them,
        # so just add them to the dictionary as nans
        if 'G_0' not in params:
            params['G_0'] = float('nan')
        if 'Gprime_0' not in params:
            params['Gprime_0'] = float('nan')

        # Check that all the required keys are in the dictionary
        expected_keys = ['V_0', 'K_0', 'Kprime_0']
        for k in expected_keys:
            if k not in params:
                raise KeyError('params object missing parameter : ' + k)

        # Finally, check that the values are reasonable.
        if params['P_0'] < 0.:
            warnings.warn('Unusual value for P_0', stacklevel=2)
        if params['V_0'] < 1.e-7 or params['V_0'] > 1.e-3:
            warnings.warn('Unusual value for V_0', stacklevel=2)
        if params['K_0'] < 1.e9 or params['K_0'] > 1.e13:
            warnings.warn('Unusual value for K_0', stacklevel=2)
        if params['Kprime_0'] < 0. or params['Kprime_0'] > 10.:
            warnings.warn('Unusual value for Kprime_0', stacklevel=2)
        if params['Kprime_prime_0'] > 0. or params['Kprime_prime_0'] < -10.:
            warnings.warn('Unusual value for Kprime_prime_0', stacklevel=2)
